#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Common Surrogate interface for IDAES.
"""
from pyomo.common.config import ConfigBlock


class SurrogateTrainer(object):
    """
    Base class for IDAES Surrogate Trainers.
    """

    CONFIG = ConfigBlock()

    def __init__(
        self,
        input_labels,
        output_labels,
        training_dataframe,
        validation_dataframe=None,
        input_bounds=None,
        **settings
    ):
        """
        This is the base class for IDAES surrogate training objects.

        Args:
           input_labels: list
              list of labels corresponding to the inputs (in order)
           output_labels: list
              list of labels corresponding to the outputs (in order)
           training_dataframe: pandas DataFrame
              Pandas DataFrame corresponding to the training data. Columns must
              include all the labels in input_labels and output_labels
           validation_dataframe: pandas DataFrame or None
             Pandas DateFrame corresponding to the validation data. Columns
             must include all the labels in input_labels and output_labels. If
             None is passed, then no validation data will be used. Some
             derived surrogate trainers may require validation data, while
             others may not.
           input_bounds: None, or dict of tuples
              if None, these are set later from the provided data
              if provided, it should be a dictionary where the keys correspond
              to the input label, and the values are tuples of bounds
              (lower, upper)
           settings: additional keyword arguments
              These are additional keyword arguments that are passed to the CONFIG
              for the derived class.
        """
        # Set the config block from passed settings
        self.config = self.CONFIG(settings)

        # We must have at least one input label and one output label
        if (
            input_labels is None
            or len(input_labels) < 1
            or output_labels is None
            or len(output_labels) < 1
        ):
            raise ValueError(
                "SurrogateTrainer requires a list of input_labels and a list "
                "of output_labels which must both have a length of at "
                "least one"
            )

        self._input_labels = list(input_labels)
        self._output_labels = list(output_labels)

        # check that the input and output labels do not overlap
        all_labels = set(self._input_labels)
        all_labels.update(self._output_labels)
        if len(all_labels) != (len(self._input_labels) + len(self._output_labels)):
            raise ValueError(
                "Duplicate label found in input_labels and/or output_labels."
            )

        # create the data members for training and validation data
        self._training_dataframe = training_dataframe
        self._validation_dataframe = validation_dataframe

        # check that all input labels and output labels are in the dataframes
        diff = set(self._input_labels) - set(self._training_dataframe.columns)
        if diff:
            raise ValueError(
                "The following input labels were not found in "
                "the training data columns: {}.".format(diff)
            )
        if self._validation_dataframe is not None:
            diff = set(self._input_labels) - set(self._validation_dataframe.columns)
            if diff:
                raise ValueError(
                    "The following input labels were not found in "
                    "the validation data columns: {}.".format(diff)
                )

        diff = set(self._output_labels) - set(self._training_dataframe.columns)
        if diff:
            raise ValueError(
                "The following output labels were not found in "
                "the training data columns: {}.".format(diff)
            )

        if self._validation_dataframe is not None:
            diff = set(self._output_labels) - set(self._validation_dataframe.columns)
            if diff:
                raise ValueError(
                    "The following output labels were not found in "
                    "the validation data columns: {}.".format(diff)
                )

        if input_bounds is not None:
            self._input_bounds = dict(input_bounds)
            # check that the labels match the input labels
            if sorted(input_bounds.keys()) != sorted(self._input_labels):
                raise ValueError(
                    "The input_labels did not match the keys in input_bounds.\n"
                    "input_bounds.keys(): {}\n"
                    "input_labels: {}".format(
                        sorted(input_bounds.keys()), sorted(input_labels)
                    )
                )
        else:
            # get the bounds from the data
            mx = self._training_dataframe.max().to_dict()
            mn = self._training_dataframe.min().to_dict()
            self._input_bounds = {k: (mn[k], mx[k]) for k in self._input_labels}

    def n_inputs(self):
        """
        The number of inputs for the surrogate

        Returns: float
        """
        return len(self._input_labels)

    def n_outputs(self):
        """
        The number of outputs for the surrogate

        Returns: float
        """
        return len(self._output_labels)

    def input_labels(self):
        """
        The ordered list of labels for the inputs

        Returns: list of strings
        """
        return list(self._input_labels)

    def output_labels(self):
        """
        The ordered list of labels for the outputs

        Returns: list of strings
        """
        return list(self._output_labels)

    def input_bounds(self):
        """
        The dictionary of input bounds. The keys of the dictionary correspond
        to the labels for the inputs. The values are tuples of (lower_bound, upper_bound)

        Returns: dict
        """
        if self._input_bounds:
            return dict(self._input_bounds)
        return None

    def train_surrogate(self):
        """
        The ``train_surrogate`` method is used to train a surrogate model
        using data provided in set_training_data. This method should return an
        instance of a derived surrogate object (from SurrogateBase)

        This method should be overridden by the derived classes.

        Returns:
           tuple : (bool, surrogate object, message) where bool indicates
           status of model training, surrogate object is an instance of a
           class derived from SurrogateBase, and message is string containing
           additional information from the trainer.

        """
        raise NotImplementedError(
            "train_surrogate called, but not implemented on the derived class"
        )


class SurrogateBase:
    """
    Base class for Surrogate object.
    """

    def __init__(self, input_labels=None, output_labels=None, input_bounds=None):
        """
        Base class for standard IDAES Surrogate object. This class is
        responsible for being able to load/save a surrogate model, evaluate
        the model given an input dataframe, and populating a block to provide
        an EO representation of the surrogate for solving in IDAES.

        Args:
           input_labels: list
              list of labels corresponding to the inputs (in order)
           output_labels: list
              list of labels corresponding to the outputs (in order)
           input_bounds: dict of tuples
              A dictionary where the keys correspond to the input label,
              and the values are tuples of bounds (lower,upper). These
              should represent the valid range for the input variables
        """
        self._input_labels = input_labels
        self._output_labels = output_labels
        self._input_bounds = input_bounds

        # check that the input and output labels do not overlap
        all_labels = set(self._input_labels)
        all_labels.update(self._output_labels)
        if len(all_labels) != (len(self._input_labels) + len(self._output_labels)):
            raise ValueError(
                "Duplicate label found in input_labels and/or output_labels."
            )

        if input_bounds is not None:
            self._input_bounds = dict(input_bounds)
            # check that the labels match the input labels
            if sorted(input_bounds.keys()) != sorted(self._input_labels):
                raise ValueError(
                    "The input_labels did not match the keys in input_bounds.\n"
                    "input_bounds.keys(): {}\n"
                    "input_labels: {}".format(
                        sorted(input_bounds.keys()), sorted(input_labels)
                    )
                )

    def n_inputs(self):
        """
        The number of inputs for the surrogate

        Returns: float
        """
        return len(self._input_labels)

    def n_outputs(self):
        """
        The number of outputs for the surrogate

        Returns: float
        """
        return len(self._output_labels)

    def input_labels(self):
        """
        The ordered list of labels for the inputs

        Returns: list of strings
        """
        return list(self._input_labels)

    def output_labels(self):
        """
        The ordered list of labels for the outputs

        Returns: list of strings
        """
        return list(self._output_labels)

    def input_bounds(self):
        """
        The dictionary of input bounds. The keys of the dictionary correspond
        to the labels for the inputs. The values are tuples of (lower_bound, upper_bound)

        Returns: dict
        """
        if self._input_bounds:
            return dict(self._input_bounds)
        return None

    def populate_block(self, block, additional_options=None):
        """
        Method to populate a Pyomo Block with surrogate model
        constraints and variables.

        Derived classes must overload this method.

        Args:
            block: Pyomo Block
               Component to be populated with constraints.
            additional_options: dict
               Additional options passed through from SurrogateBlock.build_model

        Returns:
            None
        """
        raise NotImplementedError(
            "SurrogateModel class has not implemented populate_block method."
        )

    def evaluate_surrogate(self, dataframe):
        """
        Method to method to evaluate surrogate model at a set of user
        provided values.

        Derived classes must overload this method

        Args:
           dataframe: pandas DataFrame
              The dataframe of input values to be used in the evaluation. The dataframe
              needs to contain a column corresponding to each of the input labels. Additional
              columns are fine, but are not used.

        Returns:
            output: pandas Dataframe
              Returns a dataframe of the the output values evaluated at the provided inputs.
              The index of the output dataframe should match the index of the provided inputs.
        """
        raise NotImplementedError(
            "SurrogateModel class has not implemented an evaluate_surrogate method."
        )

    def save_to_file(self, filename, overwrite=False):
        """
        This method saves an instance of the surrogate to a file so the model
        can be used later.

        Args:
           filename : str
              The path of the filename where the model will be saved
           overwrite : bool
              If True, this method will overwrite the file if it exists. If False
              and the file already exists, this will throw an error.
        """
        arg = "x"
        if overwrite:
            arg = "w"

        with open(filename, arg) as fd:
            self.save(fd)

    def save(self, strm):
        """
        Save an instance of this surrogate to the strm so the model can be used later.
        This method should be overloaded in derived surrogate classes.

        Args:
           strm: IO.TextIO
              This is the python stream like a file object or StringIO that will be used
              to serialize the surrogate object. This methods will often write a string
              of json data to the stream, but the format for derived classes need not be json.
        """
        raise NotImplementedError(
            '"save" should be implemented in the class derived from SurrogateBase'
        )

    @classmethod
    def load_from_file(cls, filename):
        """
        This method creates a new surrogate object by loading the model from the provided file.

        Args:
           filename : str
              The name of the file from which to load the model.

        Returns: an instance of the derived class or None if it failed to load
        """
        with open(filename, "r") as fd:
            return cls.load(fd)

    @classmethod
    def load(cls, strm):
        """
        Create an instance of a surrogate from a stream. This method should
        be overloaded in derived surrogate classes.

        Args:
           strm: IO.TextIO
              This is the python stream like a file object or StringIO containing
              the data required to load the surrogate. This is often, but does not
              need to be json data.

        Returns: an instance of the derived class or None if it failed to load
        """
        raise NotImplementedError(
            '"load" should be implemented in the class derived from SurrogateBase'
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Common utilities for cubic EoS property packages.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

import enum
import ctypes
from pyomo.environ import ExternalFunction, units as pyunits
from pyomo.common.fileutils import find_library

try:
    # When compiling these, I don't bother changing the extension based on OS,
    # so the file name is always ends in .so. It's fine.
    cubic_so_path = find_library("cubic_roots")
    ctypes.cdll.LoadLibrary(cubic_so_path)
except Exception:  # pylint: disable=W0703
    cubic_so_path = None


def cubic_roots_available():
    """Make sure the compiled cubic root functions are available. Yes, in
    Windows the .so extension is still used.
    """
    if cubic_so_path is None:
        return False
    return True


class CubicType(enum.Enum):
    """enum of supported cubic types."""

    PR = 0
    SRK = 1


EoS_param = {
    CubicType.PR: {"u": 2, "w": -1, "omegaA": 0.45724, "coeff_b": 0.07780},
    CubicType.SRK: {"u": 1, "w": 0, "omegaA": 0.42748, "coeff_b": 0.08664},
}


class _ExternalFunctionSpecs(object):
    """Stores the information for creating an ExternalFunction object including.
    the library, function, and units of measure for the arguments and the return
    values.
    """

    def __init__(self, func, lib, units=None, arg_units=None):
        self.func = func
        self.lib = lib
        self.units = units
        self.arg_units = arg_units

    def kwargs(self):
        return {
            "library": self.lib,
            "function": self.func,
            "units": self.units,
            "arg_units": self.arg_units,
        }


class CubicThermoExpressions(object):
    """This class provides some standard expressions for cubic equations of
    state.
    """

    _external = {
        "compress_fact_liq_func": _ExternalFunctionSpecs(
            "cubic_root_l",
            cubic_so_path,
            units=pyunits.dimensionless,
            arg_units=[
                pyunits.dimensionless,
                pyunits.dimensionless,
                pyunits.dimensionless,
            ],
        ),
        "compress_fact_vap_func": _ExternalFunctionSpecs(
            "cubic_root_h",
            cubic_so_path,
            units=pyunits.dimensionless,
            arg_units=[
                pyunits.dimensionless,
                pyunits.dimensionless,
                pyunits.dimensionless,
            ],
        ),
    }

    def __init__(self, blk):
        if not cubic_roots_available():
            raise RuntimeError("Cubic root external functions are not available.")
        self.blk = blk

    def add_funcs(self, names=None):
        if names is None:
            names = []
        elif isinstance(names, str):
            names = [names]
        for name in names:
            if hasattr(self.blk, name):
                continue
            setattr(self.blk, name, ExternalFunction(**self._external[name].kwargs()))

    def z_liq(self, eos=None, u=None, w=None, A=None, B=None, b=None, c=None, d=None):
        self.add_funcs(names="compress_fact_liq_func")
        if b is not None and c is not None and d is not None:
            pass
        else:
            if eos is not None:
                u = EoS_param[eos]["u"]
                w = EoS_param[eos]["w"]
            if u is None or w is None or A is None or B is None:
                raise RuntimeError(
                    "Please supply args: eos, {u, w, A, B}, or {b, c, d}"
                )
            b = -(1.0 + B - u * B)
            c = A + w * B**2 - u * B - u * B**2
            d = -A * B - w * B**2 - w * B**3
        return self.blk.compress_fact_liq_func(b, c, d)

    def z_vap(self, eos=None, u=None, w=None, A=None, B=None, b=None, c=None, d=None):
        self.add_funcs(names="compress_fact_vap_func")
        if b is not None and c is not None and d is not None:
            pass
        else:
            if eos is not None:
                u = EoS_param[eos]["u"]
                w = EoS_param[eos]["w"]
            if u is None or w is None or A is None or B is None:
                raise RuntimeError(
                    "Please supply args: eos, {u, w, A, B}, or {b, c, d}"
                )
            b = -(1.0 + B - u * B)
            c = A + w * B**2 - u * B - u * B**2
            d = -A * B - w * B**2 - w * B**3
        return self.blk.compress_fact_vap_func(b, c, d)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################


def isZero(x, atol):
    """Determine if a floating point number is equal to zero.

    Args:
    x (float): a number
    atol (float): absolute tolerance from zero

    Returns:
    (bool) true if the number is sufficiently close to zero
    """
    return atol > x > -atol


def areEqual(x, y, atol):
    """Determine if two floating point numbers are equal.

    Args:
    x (float): a number
    y (float): a number
    atol (float): absolute tolerance for equality

    Returns:
    (bool) true if the two numbers are sufficiently close
    """
    return isZero(x - y, atol)


def myArrayEq(x, y, atol):
    """Determine if two numpy arrays of floating point numbers are equal.

    Args:
    x (numpy.ndarray): a numpy array
    y (numpy.ndarray): a numpy array
    atol (float): absolute tolerance for equality

    Returns:
    (bool) true if the two arrays are equal
    """
    return (
        (x[0] - y[0] < atol)
        and (x[0] - y[0] > -atol)
        and (x[1] - y[1] < atol)
        and (x[1] - y[1] > -atol)
        and (x[2] - y[2] < atol)
        and (x[2] - y[2] > -atol)
    )


myPointEq = myArrayEq


def myPointsEq(x, y, atol):
    """Determine if two lists of numpy arrays are equal.

    Args:
    x (list<numpy.ndarray>): a list of numpy arrays
    y (list<numpy.ndarray>): a list of numpy arrays
    atol (float): absolute tolerance for equality

    Returns:
    (bool) true if the two arrays are equal
    """
    if len(x) != len(y):
        return False
    for i in range(len(x)):
        if not myArrayEq(x[i], y[i], atol):
            return False
    return True


def ListHasPoint(L, P, atol):
    """Determine if a list of numpy arrays contains a specific point.

    Args:
    x (list<numpy.ndarray>): a list of numpy arrays
    y (list<numpy.ndarray>): a list of numpy arrays
    atol (float): absolute tolerance for equality

    Returns:
    (bool) true if the two arrays are equal
    """
    for l in L:
        if myArrayEq(l, P, atol):
            return True
    return False



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Tests for rolling horizon helper classes
"""

import pytest
import sys
from collections import OrderedDict
import pyomo.environ as pyo
from idaes.apps.caprese.rolling import *


class TestTimeList(object):
    @pytest.mark.unit
    def test_constructor(self):
        empty_time_list = TimeList()
        assert hasattr(empty_time_list, "tolerance")
        assert len(empty_time_list) == 0

        tol = 0.49
        good_time_list = TimeList(time=[1, 2, 3], tolerance=tol)
        assert good_time_list.tolerance == tol

        with pytest.raises(ValueError) as err:
            # Error message will be checked later
            bad_time_list = TimeList(time=[3, 1, 2], tolerance=tol)

    @pytest.mark.unit
    def test_validate_time(self):
        increasing_time = [1, 2, 3]
        nonincreasing_time = [3, 1, 2]

        good_tol = 0.4999
        bad_tol = 0.5

        time_list_1 = TimeList(tolerance=good_tol)
        time_list_2 = TimeList(tolerance=bad_tol)

        assert time_list_1.validate_time(increasing_time) == increasing_time
        with pytest.raises(ValueError) as err:
            time = time_list_1.validate_time(nonincreasing_time)
            msg = "Time points must be increasing"
            assert msg in str(err)

        with pytest.raises(ValueError) as err:
            time = time_list_2.validate_time(increasing_time)
            msg = " and separated by more than twice the tolerance."
            assert msg in str(err)

    @pytest.mark.unit
    def test_is_within_bounds(self):
        empty_time_list = TimeList()

        with pytest.raises(ValueError) as err:
            empty_time_list.is_within_bounds(0.0)
            msg = "List is empty"
            assert msg in str(err)

        time = [1, 2, 3]
        t_last = time[-1]
        tol = 0.1
        eps = sys.float_info.epsilon
        root_eps = eps**0.5
        time_list = TimeList(time, tolerance=tol)
        assert time_list.is_within_bounds(2.5)
        assert time_list.is_within_bounds(t_last + tol / 2)
        assert time_list.is_within_bounds(t_last + tol)
        assert not time_list.is_within_bounds(3.5)

        assert not time_list.is_within_bounds(3.1 + root_eps)

        # Of course, we may still violate our tolerance
        # by less than machine precision.
        assert time_list.is_within_bounds(3.1 + eps / 2)

    @pytest.mark.unit
    def test_is_valid_append(self):
        time = [1, 2, 3]
        t_last = time[-1]
        tol = 0.1
        root_eps = sys.float_info.epsilon**0.5
        empty_time_list = TimeList()
        time_list = TimeList(time, tolerance=tol)
        assert empty_time_list.is_valid_append(-1)
        assert not time_list.is_valid_append(0.0)
        assert not time_list.is_valid_append(t_last)
        assert not time_list.is_valid_append(t_last + tol)
        assert time_list.is_valid_append(t_last + 2 * tol + root_eps)

    @pytest.mark.unit
    def test_append(self):
        time = [1, 2, 3]
        t_last = time[-1]
        tol = 0.1
        root_eps = sys.float_info.epsilon**0.5
        time_list = TimeList(time, tolerance=tol)

        t_new = t_last + tol
        with pytest.raises(ValueError) as err:
            time_list.append(t_new)
            msg = "Appended time values must be"
            assert msg in str(err)

        t_new = t_last + 2 * tol + root_eps
        time_list.append(t_new)
        assert time_list == time + [t_new]

    @pytest.mark.unit
    def test_validate_extend(self):
        time = [1, 2, 3]
        t_last = time[-1]
        tol = 0.1
        root_eps = sys.float_info.epsilon**0.5
        time_list = TimeList(time, tolerance=tol)

        assert time_list.validate_extend([]) == []

        new_points = [2.5, 4]
        with pytest.raises(ValueError) as err:
            time_list.validate_extend(new_points)
            msg = "First new point must not be earlier"
            assert msg in str(err)

        new_points = [t_last, 4]
        assert time_list.validate_extend(new_points) == [4]
        new_points = [t_last + tol / 2, 4]
        assert time_list.validate_extend(new_points) == [4]

        new_points = [t_last + 3 * tol / 2, 4]
        with pytest.raises(ValueError) as err:
            time_list.validate_extend(new_points)
            msg = "must be separated by more than twice the tolerance"
            assert msg in str(err)

        new_points = [3.5, 4]
        assert time_list.validate_extend(new_points) == new_points

    @pytest.mark.unit
    def test_extend(self):
        time = [1, 2, 3]
        t_last = time[-1]
        tol = 0.1
        time_list = TimeList(time, tolerance=tol)

        new_points = [4, 5]
        time_list.extend(new_points)
        assert time_list == [1, 2, 3, 4, 5]

        new_points = [5, 6, 7]
        time_list.extend(new_points)
        assert time_list == [1, 2, 3, 4, 5, 6, 7]

    @pytest.mark.unit
    def test_find_nearest_index(self):
        time = [1, 2, 3, 4, 5]
        tol = 0.1
        time_list = TimeList(time, tolerance=tol)

        assert time_list.find_nearest_index(2.5) is None
        assert time_list.find_nearest_index(2.05) == 1
        assert time_list.find_nearest_index(1.0) == 0
        assert time_list.find_nearest_index(5.05) == 4


def make_model():
    m = pyo.ConcreteModel()
    m.time = pyo.Set(initialize=range(11))
    m.space = pyo.Set(initialize=range(3))
    m.comp = pyo.Set(initialize=["a", "b"])

    m.v1 = pyo.Var(m.time, m.space)
    m.v2 = pyo.Var(m.time, m.space, m.comp)

    @m.Block(m.time, m.space)
    def b(b, t, x):
        b.v3 = pyo.Var()
        b.v4 = pyo.Var(m.comp)

    m.v1_refs = [pyo.Reference(m.v1[:, x]) for x in m.space]
    m.v2a_refs = [pyo.Reference(m.v2[:, x, "a"]) for x in m.space]
    m.v2b_refs = [pyo.Reference(m.v2[:, x, "b"]) for x in m.space]
    m.v3_refs = [pyo.Reference(m.b[:, x].v3) for x in m.space]
    m.v4a_refs = [pyo.Reference(m.b[:, x].v4["a"]) for x in m.space]
    m.v4b_refs = [pyo.Reference(m.b[:, x].v4["a"]) for x in m.space]

    for t in m.time:
        for ref in m.v1_refs:
            ref[t] = 1.0 * t
        for ref in m.v2a_refs + m.v2b_refs:
            ref[t] = 2.0 * t
        for ref in m.v3_refs:
            ref[t] = 3.0 * t
        for ref in m.v4a_refs + m.v4b_refs:
            ref[t] = 4.0 * t

    return m


class TestVectorSeries(object):
    @pytest.mark.unit
    def test_constructor(self):
        # VectorSeries is constructed from an OrderedDict with data
        # series (lists) as values. Keys are meant to be identifiers
        # of Pyomo model components, but could technically be anything.
        # To illustrate their intended use, tests use CUIDs as keys.
        m = make_model()
        data = OrderedDict(
            [
                (pyo.ComponentUID(_slice.referent), [_slice[t].value for t in m.time])
                for _slice in m.v1_refs
            ]
        )

        name = "v1"
        tol = 0.1
        history = VectorSeries(data, time=list(m.time), name=name, tolerance=tol)
        assert history.name == name
        assert type(history.time) is TimeList
        assert history.time == m.time
        assert history.time.tolerance == tol

        for _slice in m.v1_refs:
            assert pyo.ComponentUID(_slice.referent) in history

        empty_series = VectorSeries()
        assert empty_series == OrderedDict()

    @pytest.mark.unit
    def test_dim(self):
        # This test also covers consistent_dimension and
        # validate_dimension.
        m = make_model()
        data = OrderedDict(
            [
                (pyo.ComponentUID(_slice.referent), [_slice[t].value for t in m.time])
                for _slice in m.v2a_refs + m.v2b_refs
            ]
        )
        name = "v2"
        tol = 0.1
        history = VectorSeries(data, time=list(m.time), name=name, tolerance=tol)
        assert history.dim() == len(m.space * m.comp)
        assert VectorSeries().dim() == 0

        vals = [1 for _ in m.space * m.comp]
        assert history.consistent_dimension(vals)
        assert not history.consistent_dimension([1])
        assert history.validate_dimension(vals) == vals
        with pytest.raises(ValueError) as err:
            history.validate_dimension([1])
            assert "inconsistent dimension" in str(err)

    @pytest.mark.unit
    def test_len(self):
        m = make_model()
        data = OrderedDict(
            [
                (pyo.ComponentUID(_slice.referent), [_slice[t].value for t in m.time])
                for _slice in m.v3_refs
            ]
        )
        name = "v3"
        tol = 0.1
        history = VectorSeries(data, time=list(m.time), name=name, tolerance=tol)
        assert len(history) == len(m.time)
        assert len(VectorSeries()) == 0

    @pytest.mark.unit
    def test_consistent(self):
        m = make_model()
        data = OrderedDict(
            [
                (pyo.ComponentUID(_slice.referent), [_slice[t].value for t in m.time])
                for _slice in m.v3_refs
            ]
        )
        name = "v3"
        tol = 0.1
        history = VectorSeries(data, time=list(m.time), name=name, tolerance=tol)

        # These vals are consistent because this is how
        # I happened to initialize m.v3
        vals = [3.0 * m.time.last() for _ in m.space]
        bad_vals = [m.time.last() for _ in m.space]
        assert not history.consistent([])
        assert not history.consistent(bad_vals)
        assert history.consistent(vals)

        assert VectorSeries().consistent([])
        assert not VectorSeries().consistent([1])

        empty_data = OrderedDict(
            [(pyo.ComponentUID(_slice.referent), []) for _slice in m.v1_refs]
        )
        empty_series = VectorSeries(empty_data)
        vals = [1 for _ in m.space]
        assert not empty_series.consistent([])
        assert empty_series.consistent(vals)

    @pytest.mark.unit
    def test_append(self):
        m = make_model()
        data = OrderedDict(
            [
                (pyo.ComponentUID(_slice.referent), [_slice[t].value for t in m.time])
                for _slice in m.v3_refs
            ]
        )
        name = "v3"
        tol = 0.1
        history = VectorSeries(data, time=list(m.time), name=name, tolerance=tol)
        new_t = 11
        with pytest.raises(ValueError) as err:
            history.append(new_t, [1])
            assert "inconsistent dimension" in str(err)
        new_vals = [3 * new_t for _ in m.space]
        history.append(new_t, new_vals)
        last = [series[-1] for series in history.values()]
        assert last == new_vals
        assert history.time[-1] == new_t

        with pytest.raises(ValueError) as err:
            history.append(new_t, new_vals)
            msg = "Appended time values must be"
            assert msg in str(err)

    @pytest.mark.unit
    def test_extend(self):
        m = make_model()
        midpoint = len(m.time) // 2
        tlist = list(m.time)
        time_1 = tlist[0:midpoint]
        time_2 = tlist[midpoint:]

        data_1 = OrderedDict(
            [
                (pyo.ComponentUID(_slice.referent), [_slice[t].value for t in time_1])
                for _slice in m.v1_refs
            ]
        )
        data_2 = OrderedDict(
            [
                (pyo.ComponentUID(_slice.referent), [_slice[t].value for t in time_2])
                for _slice in m.v1_refs
            ]
        )
        tol = 0.1
        history_1 = VectorSeries(data_1, time_1, tolerance=tol)
        history_2 = VectorSeries(data_2, time_2, tolerance=tol)
        history_1.extend(history_2.time, history_2)

        vals = [1.0 * t for t in m.time]
        for series in history_1.values():
            assert series == vals
        assert history_1.time == m.time

        new_time = [10, 11, 13]
        new_vals = [1 * t for t in new_time]
        new_data = [new_vals for _ in history_1.values()]
        history_1.extend(new_time, new_data)

        time = list(m.time) + new_time
        time = list(OrderedDict.fromkeys(time))  # No duplicates
        vals = [1 * t for t in time]
        for series in history_1.values():
            assert series == vals
        assert history_1.time == time

        new_time = [13, 14, 15]
        new_vals = [2 * t for t in new_time]
        new_data = [new_vals for _ in history_1.values()]
        with pytest.raises(ValueError) as err:
            history_1.extend(new_time, new_data)
            msg = "data was not consistent"
            assert mst in str(err)

        with pytest.raises(ValueError) as err:
            # Assuming that history_2.dim() != 1 ...
            history_2.extend([], [[]])
            assert "inconsistent dimension" in str(err)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Basic property package for flue gas.

Main assumptions:
    - ideal gas
    - components in flue gas: O2, N2, NO, CO2, H2O, SO2
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Param,
    PositiveReals,
    Reals,
    value,
    log,
    sqrt,
    Var,
    Expression,
    check_optimal_termination,
    units as pyunits,
)
from pyomo.common.config import ConfigValue

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    Component,
    VaporPhase,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_activated_constraints,
)
from idaes.core import MaterialBalanceType, EnergyBalanceType, MaterialFlowBasis
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
from idaes.core.util import constants
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import ConfigurationError, InitializationError

# Import Python libraries
import idaes.logger as idaeslog


# Some more information about this module
__author__ = "Boiler Subsystem Team  J. Ma, M. Zamarripa, T. Burgard"
__version__ = "3"

# Set up logger
_log = idaeslog.getLogger("idaes.unit_model.properties")


@declare_process_block_class("FlueGasParameterBlock")
class FlueGasParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for flue
    gas. The ideal gas assumption is applied.

    """

    CONFIG = PhysicalParameterBlock.CONFIG()
    CONFIG.declare(
        "components",
        ConfigValue(
            default=["N2", "O2", "NO", "CO2", "H2O", "SO2"],
            domain=list,
            description="Components to include",
        ),
    )

    def build(self):
        """Add contents to the block."""
        super().build()
        self._state_block_class = FlueGasStateBlock
        _valid_comps = ["N2", "O2", "NO", "CO2", "H2O", "SO2"]

        for j in self.config.components:
            if j not in _valid_comps:
                raise ConfigurationError(f"Component '{j}' is not supported")
            self.add_component(j, Component())

        # Create Phase object
        self.Vap = VaporPhase()

        # Molecular weight
        self.mw_comp = Param(
            self.component_list,
            initialize={
                k: v
                for k, v in {
                    "O2": 0.0319988,
                    "N2": 0.0280134,
                    "NO": 0.0300061,
                    "CO2": 0.0440095,
                    "H2O": 0.0180153,
                    "SO2": 0.064064,
                }.items()
                if k in self.component_list
            },
            doc="Molecular Weight [kg/mol]",
            units=pyunits.kg / pyunits.mol,
        )

        # Thermodynamic reference state
        self.pressure_ref = Param(
            within=PositiveReals,
            default=1.01325e5,
            doc="Reference pressure [Pa]",
            units=pyunits.Pa,
        )

        self.temperature_ref = Param(
            within=PositiveReals,
            default=298.15,
            doc="Reference temperature [K]",
            units=pyunits.K,
        )

        # Critical Properties
        self.pressure_crit = Param(
            self.component_list,
            within=PositiveReals,
            initialize={
                k: v
                for k, v in {
                    "O2": 50.45985e5,
                    "N2": 33.943875e5,
                    "NO": 64.85e5,
                    "CO2": 73.8e5,
                    "H2O": 220.64e5,
                    "SO2": 7.883e6,
                }.items()
                if k in self.component_list
            },
            doc="Critical pressure [Pa]",
            units=pyunits.Pa,
        )

        self.temperature_crit = Param(
            self.component_list,
            within=PositiveReals,
            initialize={
                k: v
                for k, v in {
                    "O2": 154.58,
                    "N2": 126.19,
                    "NO": 180.0,
                    "CO2": 304.18,
                    "H2O": 647,
                    "SO2": 430.8,
                }.items()
                if k in self.component_list
            },
            doc="Critical temperature [K]",
            units=pyunits.K,
        )

        # Constants for specific heat capacity, enthalpy, and entropy
        # calculations for ideal gas (from NIST 01/08/2020
        # https://webbook.nist.gov/cgi/cbook.cgi?ID=C7727379&Units=SI&Mask=1#Thermo-Gas)
        cp_mol_ig_comp_coeff_parameter_A = {
            k: v
            for k, v in {
                "N2": 19.50583,
                "O2": 30.03235,
                "CO2": 24.99735,
                "H2O": 30.092,
                "NO": 23.83491,
                "SO2": 21.43049,
            }.items()
            if k in self.component_list
        }
        cp_mol_ig_comp_coeff_parameter_B = {
            k: v
            for k, v in {
                "N2": 19.88705,
                "O2": 8.772972,
                "CO2": 55.18696,
                "H2O": 6.832514,
                "NO": 12.58878,
                "SO2": 74.35094,
            }.items()
            if k in self.component_list
        }
        cp_mol_ig_comp_coeff_parameter_C = {
            k: v
            for k, v in {
                "N2": -8.598535,
                "O2": -3.98813,
                "CO2": -33.69137,
                "H2O": 6.793435,
                "NO": -1.139011,
                "SO2": -57.75217,
            }.items()
            if k in self.component_list
        }
        cp_mol_ig_comp_coeff_parameter_D = {
            k: v
            for k, v in {
                "N2": 1.369784,
                "O2": 0.788313,
                "CO2": 7.948387,
                "H2O": -2.53448,
                "NO": -1.497459,
                "SO2": 16.35534,
            }.items()
            if k in self.component_list
        }
        cp_mol_ig_comp_coeff_parameter_E = {
            k: v
            for k, v in {
                "N2": 0.527601,
                "O2": -0.7416,
                "CO2": -0.136638,
                "H2O": 0.082139,
                "NO": 0.214194,
                "SO2": 0.086731,
            }.items()
            if k in self.component_list
        }
        cp_mol_ig_comp_coeff_parameter_F = {
            k: v
            for k, v in {
                "N2": -4.935202,
                "O2": -11.3247,
                "CO2": -403.6075,
                "H2O": -250.881,
                "NO": 83.35783,
                "SO2": -305.7688,
            }.items()
            if k in self.component_list
        }
        cp_mol_ig_comp_coeff_parameter_G = {
            k: v
            for k, v in {
                "N2": 212.39,
                "O2": 236.1663,
                "CO2": 228.2431,
                "H2O": 223.3967,
                "NO": 237.1219,
                "SO2": 254.8872,
            }.items()
            if k in self.component_list
        }
        cp_mol_ig_comp_coeff_parameter_H = {
            k: v
            for k, v in {
                "N2": 0,
                "O2": 0,
                "CO2": -393.5224,
                "H2O": -241.8264,
                "NO": 90.29114,
                "SO2": -296.8422,
            }.items()
            if k in self.component_list
        }

        self.cp_mol_ig_comp_coeff_A = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_A,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.J / pyunits.mol / pyunits.K,
        )
        self.cp_mol_ig_comp_coeff_B = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_B,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.J / pyunits.mol / pyunits.K / pyunits.kK,
        )
        self.cp_mol_ig_comp_coeff_C = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_C,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.J / pyunits.mol / pyunits.K / pyunits.kK**2,
        )
        self.cp_mol_ig_comp_coeff_D = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_D,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.J / pyunits.mol / pyunits.K / pyunits.kK**3,
        )
        self.cp_mol_ig_comp_coeff_E = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_E,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.J / pyunits.mol / pyunits.K * pyunits.kK**2,
        )
        self.cp_mol_ig_comp_coeff_F = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_F,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.kJ / pyunits.mol,
        )
        self.cp_mol_ig_comp_coeff_G = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_G,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.J / pyunits.mol / pyunits.K,
        )
        self.cp_mol_ig_comp_coeff_H = Param(
            self.component_list,
            initialize=cp_mol_ig_comp_coeff_parameter_H,
            doc="Constants for spec. heat capacity for ideal gas",
            units=pyunits.kJ / pyunits.mol,
        )

        # Viscosity and thermal conductivity parameters
        self.ce_param = Param(
            initialize=2.6693e-5,
            units=(
                pyunits.g**0.5
                * pyunits.mol**0.5
                * pyunits.angstrom**2
                * pyunits.K**-0.5
                * pyunits.cm**-1
                * pyunits.s**-1
            ),
            doc="Parameter for the Chapman-Enskog viscosity correlation",
        )

        self.sigma = Param(
            self.component_list,
            initialize={
                k: v
                for k, v in {
                    "O2": 3.458,
                    "N2": 3.621,
                    "NO": 3.47,
                    "CO2": 3.763,
                    "H2O": 2.605,
                    "SO2": 4.29,
                }.items()
                if k in self.component_list
            },
            doc="collision diameter",
            units=pyunits.angstrom,
        )
        self.ep_Kappa = Param(
            self.component_list,
            initialize={
                k: v
                for k, v in {
                    "O2": 107.4,
                    "N2": 97.53,
                    "NO": 119.0,
                    "CO2": 244.0,
                    "H2O": 572.4,
                    "SO2": 252.0,
                }.items()
                if k in self.component_list
            },
            doc="Boltzmann constant divided by characteristic Lennard-Jones energy",
            units=pyunits.K,
        )

        self.set_default_scaling("flow_mol", 1e-4)
        self.set_default_scaling("flow_mass", 1e-3)
        self.set_default_scaling("flow_vol", 1e-3)
        # anything not explicitly listed
        self.set_default_scaling("mole_frac_comp", 1)
        self.set_default_scaling("mole_frac_comp", 1e3, index="NO")
        self.set_default_scaling("mole_frac_comp", 1e3, index="SO2")
        self.set_default_scaling("mole_frac_comp", 1e2, index="H2O")
        self.set_default_scaling("mole_frac_comp", 1e2, index="CO2")
        self.set_default_scaling("flow_vol", 1)

        # For flow_mol_comp, will calculate from flow_mol and mole_frac_comp
        # user should set a scale for both, and for each component of
        # mole_frac_comp
        self.set_default_scaling("pressure", 1e-5)
        self.set_default_scaling("temperature", 1e-1)
        self.set_default_scaling("pressure_red", 1e-3)
        self.set_default_scaling("temperature_red", 1)
        self.set_default_scaling("enth_mol_phase", 1e-3)
        self.set_default_scaling("enth_mol", 1e-3)
        self.set_default_scaling("entr_mol", 1e-2)
        self.set_default_scaling("entr_mol_phase", 1e-2)
        self.set_default_scaling("cp_mol", 0.1)
        self.set_default_scaling("cp_mol_phase", 0.1)
        self.set_default_scaling("compress_fact", 1)
        self.set_default_scaling("dens_mol_phase", 1)
        self.set_default_scaling("pressure_sat", 1e-4)
        self.set_default_scaling("visc_d_comp", 1e4)
        self.set_default_scaling("therm_cond_comp", 1e2)
        self.set_default_scaling("visc_d", 1e4)
        self.set_default_scaling("therm_cond", 1e2)
        self.set_default_scaling("mw", 1)
        self.set_default_scaling("mw_comp", 1)

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "flow_mol_comp": {"method": None, "units": "mol/s"},
                "pressure": {"method": None, "units": "Pa"},
                "temperature": {"method": None, "units": "K"},
                "pressure_crit": {"method": None, "units": "Pa"},
                "temperature_crit": {"method": None, "units": "K"},
                "pressure_red": {"method": None, "units": None},
                "temperature_red": {"method": None, "units": None},
                "enth_mol_phase": {"method": "_enthalpy_calc", "units": "J/mol"},
                "entr_mol_phase": {"method": "_entropy_calc", "units": "J/mol/K"},
                "enth_mol": {"method": "_enthalpy_calc", "units": "J/mol"},
                "entr_mol": {"method": "_entropy_calc", "units": "J/mol.K"},
                "cp_mol": {"method": "_heat_cap_calc", "units": "J/mol.K"},
                "cp_mol_phase": {"method": "_heat_cap_calc", "units": "J/mol/K"},
                "compress_fact": {"method": "_compress_fact", "units": None},
                "dens_mol_phase": {"method": "_dens_mol_phase", "units": "mol/m^3"},
                "pressure_sat": {"method": "_vapor_pressure", "units": "Pa"},
                "flow_vol": {"method": "_flow_volume", "units": "m^3/s"},
                "visc_d": {"method": "_therm_cond", "units": "kg/m-s"},
                "therm_cond": {"method": "_therm_cond", "units": "W/m-K"},
                "mw_comp": {"method": None, "units": "kg/mol"},
                "mw": {"method": None, "units": "kg/mol"},
            }
        )

        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


class _FlueGasStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(
        self,
        state_args=None,
        hold_state=False,
        state_vars_fixed=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """Initialisation routine for property package.

        Key values for the state_args dict:
            flow_mol_comp : value at which to initialize component flows
                (default=27.5e3 mol/s)
            pressure : value at which to initialize pressure
                (default=2.97e7 Pa)
            temperature : value at which to initialize temperature
                (default=866.5 K)

        Args:
            outlvl: sets logging level
            state_vars_fixed: Flag to denote state vars have been fixed.
                - True - states have been fixed by the control volume 1D.
                         Control volume 0D does not fix the state vars, so will
                         be False if this state block is used with 0D blocks.
                - False - states have not been fixed. The state block will deal
                          with fixing/unfixing.
            optarg: solver options dictionary object (default=None, use
                    default solver options)
            solver: str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state: flag indicating whether the initialization routine
                should unfix any state variables fixed during initialization
                (default=False).
                - True - states variables are not unfixed, and a dict of
                         returned containing flags for which states were fixed
                         during initialization.
                - False - state variables are unfixed after initialization by
                          calling the release_state method

            Returns:
                If hold_states is True, returns a dict containing flags for
                which states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="properties")

        # Create solver
        opt = get_solver(solver, optarg)

        if state_vars_fixed is False:
            flags = fix_state_vars(self, state_args)
        # Check when the state vars are fixed already result in dof 0
        for b in self.values():
            if degrees_of_freedom(b) != 0:
                raise InitializationError(
                    f"{self.name} State vars fixed but degrees of freedom not 0"
                )
        # ---------------------------------------------------------------------
        # Solve 1st stage
        for b in self.values():
            deactivate_list = []
            if hasattr(b, "enthalpy_correlation"):
                deactivate_list.append(b.enthalpy_correlation)
            if hasattr(b, "volumetric_flow_calculation"):
                deactivate_list.append(b.volumetric_flow_calculation)
            if hasattr(b, "entropy_correlation"):
                deactivate_list.append(b.entropy_correlation)
            for c in deactivate_list:
                c.deactivate()

            if number_activated_constraints(b) > 0:
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = opt.solve(b, tee=slc.tee)
            else:
                res = "skipped"
            init_log.info_high(
                "Initialization Step 1 {}.".format(idaeslog.condition(res))
            )

            for c in deactivate_list:
                c.activate()

            if number_activated_constraints(b) > 0:
                with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                    res = opt.solve(b, tee=slc.tee)
            else:
                res = "skipped"
            init_log.info_high(
                "Initialization Step 2 {}.".format(idaeslog.condition(res))
            )

        if not res == "skipped" and not check_optimal_termination(res):
            raise InitializationError(f"Solve failed {res}.")

        init_log.info(f"Initialisation Complete, {idaeslog.condition(res)}.")
        # ---------------------------------------------------------------------
        # If input block, return flags, else release state
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        # Unfix state variables
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        revert_state_vars(self, flags)
        init_log.info("{} State Released.".format(self.name))


@declare_process_block_class("FlueGasStateBlock", block_class=_FlueGasStateBlock)
class FlueGasStateBlockData(StateBlockData):
    """
    This is an example of a property package for calculating the thermophysical
    properties of flue gas using the ideal gas assumption.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(FlueGasStateBlockData, self).build()
        comps = self.params.component_list
        # Add state variables
        self.flow_mol_comp = Var(
            comps,
            domain=Reals,
            initialize=1.0,
            bounds=(0, 1e6),
            doc="Component molar flowrate [mol/s]",
            units=pyunits.mol / pyunits.s,
        )
        self.pressure = Var(
            domain=Reals,
            initialize=1.01325e5,
            bounds=(1, 5e7),
            doc="State pressure [Pa]",
            units=pyunits.Pa,
        )
        self.temperature = Var(
            domain=Reals,
            initialize=500,
            bounds=(200, 1500),
            doc="State temperature [K]",
            units=pyunits.K,
        )

        # Add expressions for some basic oft-used quantiies
        self.flow_mol = Expression(expr=sum(self.flow_mol_comp[j] for j in comps))

        def rule_mole_frac(b, c):
            return b.flow_mol_comp[c] / b.flow_mol

        self.mole_frac_comp = Expression(
            comps, rule=rule_mole_frac, doc="mole fraction of component i"
        )

        self.flow_mass = Expression(
            expr=sum(self.flow_mol_comp[j] * self.params.mw_comp[j] for j in comps),
            doc="total mass flow",
        )

        def rule_mw_comp(b, j):
            return b.params.mw_comp[j]

        self.mw_comp = Expression(comps, rule=rule_mw_comp)

        def rule_mw(b):
            return sum(b.mw_comp[j] * b.mole_frac_comp[j] for j in comps)

        self.mw = Expression(rule=rule_mw)

        self.pressure_crit = Expression(
            expr=sum(
                self.params.pressure_crit[j] * self.mole_frac_comp[j] for j in comps
            )
        )
        self.temperature_crit = Expression(
            expr=sum(
                self.params.temperature_crit[j] * self.mole_frac_comp[j] for j in comps
            )
        )
        self.pressure_red = Expression(expr=self.pressure / self.pressure_crit)
        self.temperature_red = Expression(expr=self.temperature / self.temperature_crit)

        self.compress_fact = Expression(expr=1.0, doc="Vapor Compressibility Factor")

        def rule_dens_mol_phase(b, p):
            return (
                b.pressure
                / b.compress_fact
                / constants.Constants.gas_constant
                / b.temperature
            )

        self.dens_mol_phase = Expression(
            self.params.phase_list, rule=rule_dens_mol_phase, doc="Molar Density"
        )

        self.flow_vol = Expression(
            doc="Volumetric Flowrate", expr=self.flow_mol / self.dens_mol_phase["Vap"]
        )

    def _heat_cap_calc(self):
        # heat capacity J/mol-K
        self.cp_mol = Var(
            initialize=1000,
            doc="heat capacity [J/mol-K]",
            units=pyunits.J / pyunits.mol / pyunits.K,
        )

        def rule_cp_phase(b, p):
            # This property module only has one phase
            return self.cp_mol

        self.cp_mol_phase = Expression(self.params.phase_list, rule=rule_cp_phase)

        try:
            ft = sum(self.flow_mol_comp[j] for j in self.params.component_list)
            t = pyunits.convert(self.temperature, to_units=pyunits.kK)
            self.heat_cap_correlation = Constraint(
                expr=(
                    self.cp_mol * ft
                    == sum(
                        self.flow_mol_comp[j]
                        * (
                            self.params.cp_mol_ig_comp_coeff_A[j]
                            + self.params.cp_mol_ig_comp_coeff_B[j] * t
                            + self.params.cp_mol_ig_comp_coeff_C[j] * t**2
                            + self.params.cp_mol_ig_comp_coeff_D[j] * t**3
                            + self.params.cp_mol_ig_comp_coeff_E[j] / t**2
                        )
                        for j in self.params.component_list
                    )
                )
            )
        except AttributeError:
            self.del_component(self.cp_mol)
            self.del_component(self.heat_cap_correlation)

    def _enthalpy_calc(self):
        self.enth_mol = Var(
            doc="Specific Enthalpy [J/mol]", units=pyunits.J / pyunits.mol
        )

        def rule_enth_phase(b, p):
            # This property module only has one phase
            return self.enth_mol

        self.enth_mol_phase = Expression(self.params.phase_list, rule=rule_enth_phase)

        def enthalpy_correlation(b):
            ft = sum(self.flow_mol_comp[j] for j in self.params.component_list)
            t = pyunits.convert(self.temperature, to_units=pyunits.kK)
            return self.enth_mol * ft == sum(
                self.flow_mol_comp[j]
                * pyunits.convert(
                    self.params.cp_mol_ig_comp_coeff_A[j] * t
                    + self.params.cp_mol_ig_comp_coeff_B[j] * t**2 / 2
                    + self.params.cp_mol_ig_comp_coeff_C[j] * t**3 / 3
                    + self.params.cp_mol_ig_comp_coeff_D[j] * t**4 / 4
                    - self.params.cp_mol_ig_comp_coeff_E[j] / t
                    + self.params.cp_mol_ig_comp_coeff_F[j],
                    to_units=pyunits.J / pyunits.mol,
                )
                for j in self.params.component_list
            )
            # NOTE: the H term (from the Shomate Equation) is not
            # included here so that the reference state enthalpy is the
            # enthalpy of formation (not 0).

        try:
            self.enthalpy_correlation = Constraint(rule=enthalpy_correlation)
        except AttributeError:
            self.del_component(self.enth_mol_phase)
            self.del_component(self.enth_mol)
            self.del_component(self.enthalpy_correlation)

    def _entropy_calc(self):
        self.entr_mol = Var(
            doc="Specific Entropy [J/mol/K]", units=pyunits.J / pyunits.mol / pyunits.K
        )
        # Specific Entropy

        def rule_entr_phase(b, p):
            # This property module only has one phase
            return self.entr_mol

        self.entr_mol_phase = Expression(self.params.phase_list, rule=rule_entr_phase)

        def entropy_correlation(b):
            ft = sum(self.flow_mol_comp[j] for j in self.params.component_list)
            t = pyunits.convert(self.temperature, to_units=pyunits.kK)
            n = self.flow_mol_comp
            x = self.mole_frac_comp
            p = self.pressure
            r_gas = constants.Constants.gas_constant
            return (self.entr_mol + r_gas * log(p / 1e5)) * ft == sum(
                n[j]
                * (
                    self.params.cp_mol_ig_comp_coeff_A[j] * log(t)
                    + self.params.cp_mol_ig_comp_coeff_B[j] * t
                    + self.params.cp_mol_ig_comp_coeff_C[j] * t**2 / 2
                    + self.params.cp_mol_ig_comp_coeff_D[j] * t**3 / 3
                    - self.params.cp_mol_ig_comp_coeff_E[j] / t**2 / 2
                    + self.params.cp_mol_ig_comp_coeff_G[j]
                    + r_gas * log(x[j])
                )
                for j in self.params.component_list
            )

        try:
            self.entropy_correlation = Constraint(rule=entropy_correlation)
        except AttributeError:
            self.del_component(self.entr_mol_phase)
            self.del_component(self.entropy_correlation)

    def _therm_cond(self):
        comps = self.params.component_list
        self.therm_cond_comp = Var(
            comps,
            initialize=0.05,
            doc="thermal conductivity J/m-K-s",
            units=pyunits.J / pyunits.m / pyunits.K / pyunits.s,
        )
        self.therm_cond = Var(
            initialize=0.05,
            doc="thermal conductivity of gas mixture J/m-K-s",
            units=pyunits.J / pyunits.m / pyunits.K / pyunits.s,
        )
        self.visc_d_comp = Var(
            comps,
            initialize=2e-5,
            doc="dynamic viscocity of pure gas species",
            units=pyunits.kg / pyunits.m / pyunits.s,
        )
        self.visc_d = Var(
            initialize=2e-5,
            doc="viscosity of gas mixture kg/m-s",
            units=pyunits.kg / pyunits.m / pyunits.s,
        )

        try:

            def rule_therm_cond(b, c):
                t = pyunits.convert(b.temperature, to_units=pyunits.kK)
                return (
                    b.therm_cond_comp[c]
                    == (
                        (
                            (
                                b.params.cp_mol_ig_comp_coeff_A[c]
                                + b.params.cp_mol_ig_comp_coeff_B[c] * t
                                + b.params.cp_mol_ig_comp_coeff_C[c] * t**2
                                + b.params.cp_mol_ig_comp_coeff_D[c] * t**3
                                + b.params.cp_mol_ig_comp_coeff_E[c] / t**2
                            )
                            / b.params.mw_comp[c]
                        )
                        + 1.25
                        * (constants.Constants.gas_constant / b.params.mw_comp[c])
                    )
                    * b.visc_d_comp[c]
                )

            self.therm_cond_con = Constraint(comps, rule=rule_therm_cond)

            def rule_theta(b, c):
                return b.temperature / b.params.ep_Kappa[c]

            self.theta = Expression(comps, rule=rule_theta)

            def rule_omega(b, c):
                return (
                    1.5794145
                    + 0.00635771 * b.theta[c]
                    - 0.7314 * log(b.theta[c])
                    + 0.2417357 * log(b.theta[c]) ** 2
                    - 0.0347045 * log(b.theta[c]) ** 3
                )

            self.omega = Expression(comps, rule=rule_omega)

            # Pure gas viscocity - from Chapman-Enskog theory
            def rule_visc_d(b, c):
                return (
                    pyunits.convert(
                        b.visc_d_comp[c], to_units=pyunits.g / pyunits.cm / pyunits.s
                    )
                    == b.params.ce_param
                    * sqrt(
                        pyunits.convert(
                            b.params.mw_comp[c], to_units=pyunits.g / pyunits.mol
                        )
                        * b.temperature
                    )
                    / b.params.sigma[c] ** 2
                    / b.omega[c]
                )

            self.visc_d_con = Constraint(comps, rule=rule_visc_d)

            # section to calculate viscosity of gas mixture
            def rule_phi(b, i, j):
                return (
                    1
                    / 2.8284
                    * (1 + (b.params.mw_comp[i] / b.params.mw_comp[j])) ** (-0.5)
                    * (
                        1
                        + sqrt(b.visc_d_comp[i] / b.visc_d_comp[j])
                        * (b.params.mw_comp[j] / b.params.mw_comp[i]) ** 0.25
                    )
                    ** 2
                )

            self.phi_ij = Expression(comps, comps, rule=rule_phi)

            # viscosity of Gas mixture kg/m-s
            def rule_visc_d_mix(b):
                return b.visc_d == sum(
                    (b.mole_frac_comp[i] * b.visc_d_comp[i])
                    / sum(b.mole_frac_comp[j] * b.phi_ij[i, j] for j in comps)
                    for i in comps
                )

            self.vis_d_mix_con = Constraint(rule=rule_visc_d_mix)

            # thermal conductivity of gas mixture in kg/m-s
            def rule_therm_mix(b):
                return b.therm_cond == sum(
                    (b.mole_frac_comp[i] * b.therm_cond_comp[i])
                    / sum(b.mole_frac_comp[j] * b.phi_ij[i, j] for j in comps)
                    for i in comps
                )

            self.therm_mix_con = Constraint(rule=rule_therm_mix)

        except AttributeError:
            self.del_component(self.therm_cond_comp)
            self.del_component(self.therm_cond)
            self.del_component(self.visc_d_comp)
            self.del_component(self.visc_d)
            self.del_component(self.omega)
            self.del_component(self.theta)
            self.del_component(self.phi_ij)
            self.del_component(self.sigma)
            self.del_component(self.ep_Kappa)
            self.del_component(self.therm_cond_con)
            self.del_component(self.theta_con)
            self.del_component(self.omega_con)
            self.del_component(self.visc_d_con)
            self.del_component(self.phi_con)

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_terms(self, p, j):
        return self.flow_mol_comp[j]

    def get_material_flow_basis(self):
        return MaterialFlowBasis.molar

    def get_enthalpy_flow_terms(self, p):
        if not self.is_property_constructed("enthalpy_flow_terms"):
            try:

                def rule_enthalpy_flow_terms(b, p):
                    return self.enth_mol_phase[p] * self.flow_mol

                self.enthalpy_flow_terms = Expression(
                    self.params.phase_list, rule=rule_enthalpy_flow_terms
                )
            except AttributeError:
                self.del_component(self.enthalpy_flow_terms)
        return self.enthalpy_flow_terms[p]

    def get_material_density_terms(self, p, j):
        return self.dens_mol_phase[p]

    def get_energy_density_terms(self, p):
        if not self.is_property_constructed("energy_density_terms"):
            try:

                def rule_energy_density_terms(b, p):
                    return (
                        self.enth_mol_phase[p] * self.dens_mol_phase[p] - self.pressure
                    )

                self.energy_density_terms = Expression(
                    self.params.phase_list, rule=rule_energy_density_terms
                )
            except AttributeError:
                self.del_component(self.energy_density_terms)
        return self.energy_density_terms[p]

    def define_state_vars(self):
        return {
            "flow_mol_comp": self.flow_mol_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    def model_check(self):
        """
        Model checks for property block
        """
        # Check temperature bounds
        for v in self.component_object_data(Var, descend_into=True):
            if value(v) < v.lb:
                _log.error(f"{v} is below lower bound in {self.name}")
            if value(v) > v.ub:
                _log.error(f"{v} is above upper bound in {self.name}")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # Get some scale factors that are frequently used to calculate others
        sf_flow = iscale.get_scaling_factor(self.flow_mol)
        sf_mol_fraction = {}
        comps = self.params.component_list
        for i in comps:
            sf_mol_fraction[i] = iscale.get_scaling_factor(self.mole_frac_comp[i])
        # calculate flow_mol_comp scale factors
        for i, c in self.flow_mol_comp.items():
            iscale.set_scaling_factor(c, sf_flow * sf_mol_fraction[i])

        if self.is_property_constructed("energy_density_terms"):
            for i, c in self.energy_density_terms.items():
                sf1 = iscale.get_scaling_factor(self.enth_mol_phase[i])
                sf2 = iscale.get_scaling_factor(self.dens_mol_phase[i])
                iscale.set_scaling_factor(c, sf1 * sf2)

        if self.is_property_constructed("enthalpy_flow_terms"):
            for i, c in self.enthalpy_flow_terms.items():
                sf1 = iscale.get_scaling_factor(self.enth_mol_phase[i])
                sf2 = iscale.get_scaling_factor(self.flow_mol)
                iscale.set_scaling_factor(c, sf1 * sf2)

        if self.is_property_constructed("heat_cap_correlation"):
            iscale.constraint_scaling_transform(
                self.heat_cap_correlation,
                iscale.get_scaling_factor(self.cp_mol)
                * iscale.get_scaling_factor(self.flow_mol),
                overwrite=False,
            )
        if self.is_property_constructed("enthalpy_correlation"):
            for c in self.enthalpy_correlation.values():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.enth_mol)
                    * iscale.get_scaling_factor(self.flow_mol),
                    overwrite=False,
                )
        if self.is_property_constructed("entropy_correlation"):
            iscale.constraint_scaling_transform(
                self.entropy_correlation,
                iscale.get_scaling_factor(self.entr_mol)
                * iscale.get_scaling_factor(self.flow_mol),
                overwrite=False,
            )
        if self.is_property_constructed("vapor_pressure_correlation"):
            iscale.constraint_scaling_transform(
                self.vapor_pressure_correlation,
                log(iscale.get_scaling_factor(self.pressure_sat))
                * iscale.get_scaling_factor(self.flow_mol),
                overwrite=False,
            )
        if self.is_property_constructed("therm_cond_con"):
            for i, c in self.therm_cond_con.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.therm_cond_comp[i]),
                    overwrite=False,
                )
        if self.is_property_constructed("therm_mix_con"):
            iscale.constraint_scaling_transform(
                self.therm_mix_con,
                iscale.get_scaling_factor(self.therm_cond),
                overwrite=False,
            )
        if self.is_property_constructed("visc_d_con"):
            for i, c in self.visc_d_con.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.visc_d_comp[i]), overwrite=False
                )
            for i, c in self.vis_d_mix_con.items():
                iscale.constraint_scaling_transform(
                    self.vis_d_mix_con,
                    iscale.get_scaling_factor(self.visc_d),
                    overwrite=False,
                )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for fixed bed TSA 0D model.
"""

__author__ = "Alex Noring"

import pytest
import numpy as np

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Expression,
    Var,
    value,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.initialization import InitializationStatus
from idaes.core.solvers import get_solver

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.models.properties import iapws95
from idaes.models_extra.power_generation.properties import FlueGasParameterBlock
from idaes.models_extra.temperature_swing_adsorption import (
    FixedBedTSA0D,
    FixedBedTSA0DInitializer,
    Adsorbent,
    SteamCalculationType,
    TransformationScheme,
)
from idaes.models_extra.temperature_swing_adsorption.util import tsa_summary
import idaes.core.util.scaling as iscale

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestTsaZeolite:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.unit = FixedBedTSA0D(
            adsorbent=Adsorbent.zeolite_13x,
            number_of_beds=120,
            transformation_method="dae.collocation",
            transformation_scheme=TransformationScheme.lagrangeRadau,
            finite_elements=20,
            collocation_points=6,
            compressor=False,
            steam_calculation=SteamCalculationType.none,
        )

        m.fs.unit.inlet.flow_mol_comp[0, "H2O"].fix(0)
        m.fs.unit.inlet.flow_mol_comp[0, "CO2"].fix(40)
        m.fs.unit.inlet.flow_mol_comp[0, "N2"].fix(99960)
        m.fs.unit.inlet.flow_mol_comp[0, "O2"].fix(0)
        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(100000)

        m.fs.unit.temperature_desorption.fix(470)
        m.fs.unit.temperature_adsorption.fix(310)
        m.fs.unit.temperature_heating.fix(500)
        m.fs.unit.temperature_cooling.fix(300)
        m.fs.unit.bed_diameter.fix(4)
        m.fs.unit.bed_height.fix(8)
        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):
        assert hasattr(model.fs.unit, "heterogeneity_parameter_ref")
        assert not hasattr(model.fs.unit, "lower_saturtion_capacity")
        assert not hasattr(model.fs.unit, "toth_constant_alpha")
        assert not hasattr(model.fs.unit, "compressor")
        assert not hasattr(model.fs.unit, "flow_mass_steam")
        assert not hasattr(model.fs.unit, "steam_heater")
        assert isinstance(model.fs.unit.velocity_in, Expression)

        assert number_variables(model) == 2838
        assert number_total_constraints(model) == 2815
        assert number_unused_variables(model) == 11

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_units(self, model):
        assert_units_consistent(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initializer = FixedBedTSA0DInitializer()
        initializer.initialize(
            model.fs.unit, heating_time_guess=70000, cooling_time_guess=110000
        )
        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # CO2 rich outlet
        assert pytest.approx(2.34462, abs=1e-3) == value(
            model.fs.unit.co2_rich_stream.flow_mol_comp[0, "CO2"]
        )
        assert pytest.approx(62.07595, abs=1e-3) == value(
            model.fs.unit.co2_rich_stream.flow_mol_comp[0, "N2"]
        )
        assert pytest.approx(470, abs=1e-3) == value(
            model.fs.unit.co2_rich_stream.temperature[0]
        )
        assert pytest.approx(100000, abs=1e-3) == value(
            model.fs.unit.co2_rich_stream.pressure[0]
        )
        # N2 rich outlet
        assert pytest.approx(37.65537, abs=1e-3) == value(
            model.fs.unit.n2_rich_stream.flow_mol_comp[0, "CO2"]
        )
        assert pytest.approx(99897.92404, abs=1e-3) == value(
            model.fs.unit.n2_rich_stream.flow_mol_comp[0, "N2"]
        )
        assert pytest.approx(310, abs=1e-3) == value(
            model.fs.unit.n2_rich_stream.temperature[0]
        )
        assert pytest.approx(100000, abs=1e-3) == value(
            model.fs.unit.n2_rich_stream.pressure[0]
        )
        # air velocity
        assert pytest.approx(2.44178, abs=1e-3) == value(model.fs.unit.velocity_in)

    @pytest.mark.unit
    def test_report(self, model):
        stream_table_df = model.fs.unit._get_stream_table_contents()
        expected_columns = ["Inlet", "CO2 Rich Stream", "N2 Rich Stream", "H2O Stream"]
        for column in expected_columns:
            assert column in stream_table_df.columns
            assert not np.isnan(stream_table_df[column]["temperature"])
            assert not np.isnan(stream_table_df[column]["pressure"])

        performance_dict = model.fs.unit._get_performance_contents()
        for k in performance_dict["vars"]:
            assert performance_dict["vars"][k] is not None

    @pytest.mark.unit
    def test_summary(self, model):
        summary_df = tsa_summary(model.fs.unit)
        assert "Adsorption temperature [K]" in summary_df.index
        assert "Cycle time [h]" in summary_df.index
        assert "Pressure drop [Pa]" in summary_df.index


class TestTsaMgmof:
    # also testing calculate beds and simple steam calcs
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.unit = FixedBedTSA0D(
            adsorbent=Adsorbent.mmen_mg_mof_74,
            number_of_beds=None,
            transformation_method="dae.collocation",
            transformation_scheme=TransformationScheme.lagrangeRadau,
            finite_elements=20,
            collocation_points=6,
            compressor=False,
            steam_calculation=SteamCalculationType.simplified,
        )

        m.fs.unit.inlet.flow_mol_comp[0, "H2O"].fix(0)
        m.fs.unit.inlet.flow_mol_comp[0, "CO2"].fix(0.00960 * 0.12)
        m.fs.unit.inlet.flow_mol_comp[0, "N2"].fix(0.00960 * 0.88)
        m.fs.unit.inlet.flow_mol_comp[0, "O2"].fix(0)
        m.fs.unit.inlet.temperature.fix(300)
        m.fs.unit.inlet.pressure.fix(1e5)

        m.fs.unit.temperature_desorption.fix(430)
        m.fs.unit.temperature_adsorption.fix(310)
        m.fs.unit.temperature_heating.fix(440)
        m.fs.unit.temperature_cooling.fix(300)
        m.fs.unit.bed_diameter.fix(3 / 100)
        m.fs.unit.bed_height.fix(1.2)
        m.fs.unit.pressure_drop.fix(-4000)
        iscale.calculate_scaling_factors(m)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):
        assert not hasattr(model.fs.unit, "heterogeneity_parameter_ref")
        assert hasattr(model.fs.unit, "lower_saturtion_capacity")
        assert not hasattr(model.fs.unit, "toth_constant_alpha")
        assert not hasattr(model.fs.unit, "compressor")
        assert hasattr(model.fs.unit, "flow_mass_steam")
        assert not hasattr(model.fs.unit, "steam_heater")
        assert isinstance(model.fs.unit.velocity_in, Var)

        assert number_variables(model) == 2840
        assert number_total_constraints(model) == 2816
        assert number_unused_variables(model) == 11

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_units(self, model):
        assert_units_consistent(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initializer = FixedBedTSA0DInitializer(constraint_tolerance=1e-4)
        initializer.initialize(
            model.fs.unit, heating_time_guess=2000, cooling_time_guess=900
        )
        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # CO2 rich outlet
        assert pytest.approx(1.1486e-3, abs=1e-6) == value(
            model.fs.unit.co2_rich_stream.flow_mol_comp[0, "CO2"]
        )
        assert pytest.approx(1.6541e-5, abs=1e-8) == value(
            model.fs.unit.co2_rich_stream.flow_mol_comp[0, "N2"]
        )
        assert pytest.approx(430, abs=1e-1) == value(
            model.fs.unit.co2_rich_stream.temperature[0]
        )
        assert pytest.approx(100000, abs=1e-1) == value(
            model.fs.unit.co2_rich_stream.pressure[0]
        )
        # N2 rich outlet
        assert pytest.approx(3.3229e-6, abs=1e-9) == value(
            model.fs.unit.n2_rich_stream.flow_mol_comp[0, "CO2"]
        )
        assert pytest.approx(8.4314e-3, abs=1e-6) == value(
            model.fs.unit.n2_rich_stream.flow_mol_comp[0, "N2"]
        )
        assert pytest.approx(310, abs=1e-1) == value(
            model.fs.unit.n2_rich_stream.temperature[0]
        )
        assert pytest.approx(100000, abs=1e-1) == value(
            model.fs.unit.n2_rich_stream.pressure[0]
        )
        # air velocity
        assert pytest.approx(0.41747, abs=1e-4) == value(model.fs.unit.velocity_in)


class TestTsaPolystyrene:
    # also testing compressor and rigorous steam calcs
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.compressor_props = FlueGasParameterBlock(components=["N2", "CO2"])
        m.fs.steam_props = iapws95.Iapws95ParameterBlock()

        m.fs.unit = FixedBedTSA0D(
            adsorbent=Adsorbent.polystyrene_amine,
            number_of_beds=1,
            transformation_method="dae.collocation",
            transformation_scheme=TransformationScheme.lagrangeRadau,
            finite_elements=20,
            collocation_points=6,
            compressor=True,
            compressor_properties=m.fs.compressor_props,
            steam_calculation=SteamCalculationType.rigorous,
            steam_properties=m.fs.steam_props,
        )

        m.fs.unit.inlet.flow_mol_comp[0, "H2O"].fix(0)
        m.fs.unit.inlet.flow_mol_comp[0, "CO2"].fix(0.00960 * 0.12)
        m.fs.unit.inlet.flow_mol_comp[0, "N2"].fix(0.00960 * 0.88)
        m.fs.unit.inlet.flow_mol_comp[0, "O2"].fix(0)
        m.fs.unit.inlet.temperature.fix(300)
        m.fs.unit.inlet.pressure.fix(1e5)

        m.fs.unit.temperature_desorption.fix(445)
        m.fs.unit.temperature_adsorption.fix(440)
        m.fs.unit.temperature_heating.fix(450)
        m.fs.unit.temperature_cooling.fix(435)
        m.fs.unit.bed_diameter.fix(10 / 100)
        m.fs.unit.bed_height.fix(1.2)
        m.fs.unit.compressor.unit.efficiency_isentropic.fix(0.8)
        iscale.calculate_scaling_factors(m)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):
        assert not hasattr(model.fs.unit, "heterogeneity_parameter_ref")
        assert not hasattr(model.fs.unit, "lower_saturtion_capacity")
        assert hasattr(model.fs.unit, "toth_constant_alpha")
        assert hasattr(model.fs.unit, "compressor")
        assert hasattr(model.fs.unit, "flow_mass_steam")
        assert hasattr(model.fs.unit, "steam_heater")
        assert isinstance(model.fs.unit.velocity_in, Expression)

        assert number_variables(model) == 2869
        assert number_total_constraints(model) == 2842
        assert number_unused_variables(model) == 11

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initializer = FixedBedTSA0DInitializer()
        initializer.initialize(
            model.fs.unit, heating_time_guess=700, cooling_time_guess=700
        )
        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # CO2 rich outlet
        assert pytest.approx(1.04099e-5, abs=1e-8) == value(
            model.fs.unit.co2_rich_stream.flow_mol_comp[0, "CO2"]
        )
        assert pytest.approx(6.24804e-5, abs=1e-8) == value(
            model.fs.unit.co2_rich_stream.flow_mol_comp[0, "N2"]
        )
        assert pytest.approx(445, abs=1e-1) == value(
            model.fs.unit.co2_rich_stream.temperature[0]
        )
        assert pytest.approx(100000, abs=1e-1) == value(
            model.fs.unit.co2_rich_stream.pressure[0]
        )
        # N2 rich outlet
        assert pytest.approx(1.14159e-3, abs=1e-6) == value(
            model.fs.unit.n2_rich_stream.flow_mol_comp[0, "CO2"]
        )
        assert pytest.approx(8.3855e-3, abs=1e-6) == value(
            model.fs.unit.n2_rich_stream.flow_mol_comp[0, "N2"]
        )
        assert pytest.approx(440, abs=1e-1) == value(
            model.fs.unit.n2_rich_stream.temperature[0]
        )
        assert pytest.approx(100000, abs=1e-1) == value(
            model.fs.unit.n2_rich_stream.pressure[0]
        )
        # air velocity
        assert pytest.approx(0.06388, abs=1e-4) == value(model.fs.unit.velocity_in)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Watertank model
The water tank has only one inlet and one outlet

main assumptions:

1.- Heat loss is a variable given by the user (zero heat loss can be
specified if adiabatic)
2.- Calculate pressure change due to gravity based on water level
and contraction to downcomer
3.- Water level is either fixed for steady-state model or calculated for
dynamic model
4.- Assume enthalpy_in = enthalpy_out + heat loss

5.- Subcooled water from economizer and saturated water from waterwall
are mixed before entering the tank

Created: November 04 2020
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.environ import value, Var, Reference, acos
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.solvers import get_solver
from idaes.core.util.constants import Constants as const

import idaes.logger as idaeslog

__author__ = "Boiler Subsystem Team (J. Ma, D. Caballero, M. Zamarripa)"
__version__ = "2.0.0"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("WaterTank")
class WaterTankData(UnitModelBlockData):
    """
    Water Tank Unit Operation Class
    """

    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare(
        "tank_type",
        ConfigValue(
            default="simple_tank",
            domain=In(
                [
                    "simple_tank",
                    "rectangular_tank",
                    "vertical_cylindrical_tank",
                    "horizontal_cylindrical_tank",
                ]
            ),
            description="Flag indicating the tank type",
            doc="""Flag indicating the type of tank to be modeled, and
then calculate the volume of the filled level consequently,
**default** - simple_tank.
**Valid values:** {
**simple_tank** - use a general tank and provide the area,
**rectangular_tank** - use a rectangular tank and provide the width and length,
**vertical_cylindrical_tank** - use a vertical cylindrical tank
and provide the diameter,
**horizontal_cylindrical_tank** - use a horizontal cylindrical tank and
provide the length and diameter.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentPhase,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of material balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single ethalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - ethalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model (pre-DAE transformation).
        Args:
            None
        Returns:
            None
        """

        # Call UnitModel.build to setup dynamics
        super().build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_geometry()

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer,
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        # Add Inlet and Outlet Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add object references
        self.volume = Reference(self.control_volume.volume)

        # Set references to balance terms at unit level
        if (
            self.config.has_heat_transfer is True
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            self.heat_duty = Reference(self.control_volume.heat)

        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != "none"
        ):
            self.deltaP = Reference(self.control_volume.deltaP)

        # Set Unit Geometry and Holdup Volume
        self._set_geometry()

        # Construct performance equations
        self._make_performance()

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary
        """
        if self.config.tank_type == "simple_tank":
            # Declare a variable for cross sectional area
            self.tank_cross_sect_area = Var(
                initialize=1.0, doc="Cross-sectional" " area of the tank"
            )

        elif self.config.tank_type == "rectangular_tank":
            # Declare variables for width and length
            self.tank_width = Var(initialize=1.0, doc="Width of the tank")
            self.tank_length = Var(initialize=1.0, doc="Length of the tank")

        elif (
            self.config.tank_type == "horizontal_cylindrical_tank"
            or "vertical_cylindrical_tank"
        ):
            # Declare a variable for diameter of the tank
            self.tank_diameter = Var(initialize=0.5, doc="Inside diameter of the tank")
            if self.config.tank_type == "horizontal_cylindrical_tank":
                # Declare a variable for length of the tank
                self.tank_length = Var(initialize=1, doc="Length of the tank")

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model
        """

        # Add performance variables
        self.tank_level = Var(
            self.flowsheet().time, initialize=1.0, doc="Water level from in the tank"
        )

        # Auxiliary expressions for volume
        # Rectangular tank
        if self.config.tank_type == "rectangular_tank":
            # Calculation of cross-sectional area of the rectangle
            @self.Expression(doc="Cross-sectional area of the tank")
            def tank_cross_sect_area(b):
                return b.tank_width * b.tank_length

        # Vertical cylindrical tank
        elif self.config.tank_type == "vertical_cylindrical_tank":

            @self.Expression(doc="Radius of the tank")
            def tank_radius(b):
                return b.tank_diameter / 2

            # Calculation of cross-sectional area of the vertical cylinder
            @self.Expression(doc="Cross-sectional area of the tank")
            def tank_cross_sect_area(b):
                return const.pi * b.tank_radius**2

        # Horizontal cylindrical tank
        elif self.config.tank_type == "horizontal_cylindrical_tank":
            # Calculation of area covered by the liquid level
            # at one end of the tank
            @self.Expression(doc="Radius of the tank")
            def tank_radius(b):
                return b.tank_diameter / 2

            # Angle of the circular sector used to calculate the area

            @self.Expression(
                self.flowsheet().time,
                doc="Angle of the circular" " sector of liquid level",
            )
            def alpha_tank(b, t):
                return 2 * acos((b.tank_radius - b.tank_level[t]) / b.tank_radius)

            @self.Expression(
                self.flowsheet().time,
                doc="Area covered by the liquid level" " at one end of the tank",
            )
            def tank_area(b, t):
                return (
                    0.5 * b.alpha_tank[t] * b.tank_radius**2
                    - (b.tank_radius - b.tank_level[t])
                    * (2 * b.tank_radius * b.tank_level[t] - b.tank_level[t] ** 2)
                    ** 0.5
                )

        # Constraint for volume of the liquid in tank
        @self.Constraint(self.flowsheet().time, doc="volume of liquid in the tank")
        def volume_eqn(b, t):
            if self.config.tank_type == "horizontal_cylindrical_tank":
                return b.volume[t] == b.tank_length * b.tank_area[t]
            else:
                return b.volume[t] == b.tank_level[t] * b.tank_cross_sect_area

        # Pressure change equation due gravity
        @self.Constraint(self.flowsheet().time, doc="pressure drop")
        def pressure_change_eqn(b, t):
            return (
                b.deltaP[t]
                == b.control_volume.properties_in[t].dens_mass_phase["Liq"]
                * const.acceleration_gravity
                * b.tank_level[t]
            )

    def set_initial_condition(self):
        if self.config.dynamic is True:
            self.control_volume.material_accumulation[:, :, :].value = 0
            self.control_volume.energy_accumulation[:, :].value = 0
            self.control_volume.material_accumulation[0, :, :].fix(0)
            self.control_volume.energy_accumulation[0, :].fix(0)

    def initialize_build(
        blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        Water tank initialization routine.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) for the control_volume of the model to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = return solver state for each step in subroutines
                     * 3 = include solver output information (tee=True)

            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        init_log.info_low("Starting initialization...")

        flags = blk.control_volume.initialize(
            state_args=state_args, outlvl=outlvl, optarg=optarg, solver=solver
        )
        init_log.info_high("Initialization Step 1 Complete.")

        # Fix outlet pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.fix(
                value(blk.control_volume.properties_in[t].pressure)
            )
        blk.pressure_change_eqn.deactivate()

        # solve model
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Unfix outlet pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].pressure.unfix()
        blk.pressure_change_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        blk.control_volume.release_state(flags, outlvl)
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for ControlVolumeBlockData, and for initializing the 0D fixed bed module

Author: Brandon Paul
"""

import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    TransformationFactory,
    value,
    units as pyunits,
    Constraint,
)
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock, EnergyBalanceType
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
    unused_variables_set,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.initialization import initialize_by_time_element
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError

# Import FixedBed0D unit model
from idaes.models_extra.gas_solid_contactors.unit_models.fixed_bed_0D import FixedBed0D

# Import property packages
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.gas_phase_thermo import (
    GasPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.solid_phase_thermo import (
    SolidPhaseParameterBlock,
)
from idaes.models_extra.gas_solid_contactors.properties.methane_iron_OC_reduction.hetero_reactions import (
    HeteroReactionParameterBlock,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 3600], time_units=pyunits.s)

    # Set up thermo props and reaction props
    m.fs.gas_props = GasPhaseParameterBlock()
    m.fs.solid_props = SolidPhaseParameterBlock()
    m.fs.solid_rxns = HeteroReactionParameterBlock(
        solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
    )

    m.fs.unit = FixedBed0D(
        gas_property_package=m.fs.gas_props,
        solid_property_package=m.fs.solid_props,
        reaction_package=m.fs.solid_rxns,
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 9

    assert m.fs.unit.config.dynamic is True
    assert m.fs.unit.config.has_holdup is True
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.enthalpyTotal
    assert m.fs.unit.config.gas_property_package is m.fs.gas_props
    assert m.fs.unit.config.gas_property_package_args is None
    assert m.fs.unit.config.solid_property_package is m.fs.solid_props
    assert m.fs.unit.config.solid_property_package_args is None
    assert m.fs.unit.config.reaction_package is m.fs.solid_rxns


# -----------------------------------------------------------------------------
class TestIronOC(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 3600], time_units=pyunits.s)

        m.fs.gas_props = GasPhaseParameterBlock()
        m.fs.solid_props = SolidPhaseParameterBlock()
        m.fs.solid_rxns = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
        )

        m.fs.unit = FixedBed0D(
            gas_property_package=m.fs.gas_props,
            solid_property_package=m.fs.solid_props,
            reaction_package=m.fs.solid_rxns,
        )

        # Discretize time domain
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=100, wrt=m.fs.time, scheme="BACKWARD")

        # Set reactor design conditions
        m.fs.unit.bed_diameter.fix(1)  # diameter of the TGA reactor [m]
        m.fs.unit.bed_height.fix(1)  # height of solids in the TGA reactor [m]

        # Set initial conditions of the solid phase
        m.fs.unit.solids[0].temperature.fix(1273.15)
        m.fs.unit.solids[0].particle_porosity.fix(0.20)
        m.fs.unit.solids[0].mass_frac_comp["Fe2O3"].fix(0.45)
        m.fs.unit.solids[0].mass_frac_comp["Fe3O4"].fix(0)
        m.fs.unit.solids[0].mass_frac_comp["Al2O3"].fix(0.55)

        # Set conditions of the gas phase (this is all fixed as gas side
        # assumption is excess gas flowrate which means all state variables
        # remain unchanged)
        for t in m.fs.time:
            m.fs.unit.gas[t].temperature.fix(1273.15)
            m.fs.unit.gas[t].pressure.fix(1.01325e5)  # 1atm
            m.fs.unit.gas[t].mole_frac_comp["CO2"].fix(0.4)
            m.fs.unit.gas[t].mole_frac_comp["H2O"].fix(0.5)
            m.fs.unit.gas[t].mole_frac_comp["CH4"].fix(0.1)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert isinstance(iron_oc.fs.unit.volume_bed_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.volume_solid_constraint, Constraint)
        assert isinstance(
            iron_oc.fs.unit.solids_material_holdup_constraints, Constraint
        )
        assert isinstance(
            iron_oc.fs.unit.solids_material_accumulation_constraints, Constraint
        )
        assert isinstance(iron_oc.fs.unit.mass_solids_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.sum_component_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.solids_energy_holdup_constraints, Constraint)
        assert isinstance(
            iron_oc.fs.unit.solids_energy_accumulation_constraints, Constraint
        )

        assert number_variables(iron_oc) == 3650
        assert number_total_constraints(iron_oc) == 2925
        assert number_unused_variables(iron_oc) == 206

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.fixture(scope="class")
    def iron_oc_unscaled(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 360], time_units=pyunits.s)

        m.fs.gas_props = GasPhaseParameterBlock()
        m.fs.solid_props = SolidPhaseParameterBlock()
        m.fs.solid_rxns = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
        )

        m.fs.unit = FixedBed0D(
            gas_property_package=m.fs.gas_props,
            solid_property_package=m.fs.solid_props,
            reaction_package=m.fs.solid_rxns,
        )

        # Discretize time domain
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=10, wrt=m.fs.time, scheme="BACKWARD")

        # Set reactor design conditions
        m.fs.unit.bed_diameter.fix(1)  # diameter of the TGA reactor [m]
        m.fs.unit.bed_height.fix(1)  # height of solids in the TGA reactor [m]

        # Set initial conditions of the solid phase
        m.fs.unit.solids[0].temperature.fix(1273.15)
        m.fs.unit.solids[0].particle_porosity.fix(0.20)
        m.fs.unit.solids[0].mass_frac_comp["Fe2O3"].fix(0.45)
        m.fs.unit.solids[0].mass_frac_comp["Fe3O4"].fix(0)
        m.fs.unit.solids[0].mass_frac_comp["Al2O3"].fix(0.55)

        # Set conditions of the gas phase (this is all fixed as gas side
        # assumption is excess gas flowrate which means all state variables
        # remain unchanged)
        for t in m.fs.time:
            m.fs.unit.gas[t].temperature.fix(1273.15)
            m.fs.unit.gas[t].pressure.fix(1.01325e5)  # 1atm
            m.fs.unit.gas[t].mole_frac_comp["CO2"].fix(0.4)
            m.fs.unit.gas[t].mole_frac_comp["H2O"].fix(0.5)
            m.fs.unit.gas[t].mole_frac_comp["CH4"].fix(0.1)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_unscaled(self, iron_oc_unscaled):
        optarg = {"tol": 1e-6}

        initialization_tester(iron_oc_unscaled, optarg=optarg)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_by_time_unscaled(self, iron_oc_unscaled):

        optarg = {"tol": 1e-6}
        solver = get_solver("ipopt", optarg)  # create solver

        initialize_by_time_element(
            iron_oc_unscaled.fs, iron_oc_unscaled.fs.time, solver=solver
        )

        assert degrees_of_freedom(iron_oc_unscaled) == 0

        # Assert that model is still fixed and deactivated as expected
        assert iron_oc_unscaled.fs.unit.solids[
            iron_oc_unscaled.fs.time.first()
        ].particle_porosity.fixed

        for t in iron_oc_unscaled.fs.time:
            if t != iron_oc_unscaled.fs.time.first():
                assert not iron_oc_unscaled.fs.unit.solids[t].particle_porosity.fixed
            assert iron_oc_unscaled.fs.unit.solids[t].density_particle_constraint.active
            assert iron_oc_unscaled.fs.unit.solids[t].density_skeletal_constraint.active

        # Assert that constraints are feasible after initialization
        for con in iron_oc_unscaled.fs.component_data_objects(Constraint, active=True):
            assert value(con.body) - value(con.upper) < 1e-4
            assert value(con.lower) - value(con.body) < 1e-4

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_unscaled(self, iron_oc_unscaled):
        results = solver.solve(iron_oc_unscaled)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_scaling(self, iron_oc):

        # Set scaling gas phase for state variables
        iron_oc.fs.gas_props.set_default_scaling("flow_mol", 1e-3)
        iron_oc.fs.gas_props.set_default_scaling("pressure", 1e-5)
        iron_oc.fs.gas_props.set_default_scaling("temperature", 1e-2)
        for comp in iron_oc.fs.gas_props.component_list:
            iron_oc.fs.gas_props.set_default_scaling("mole_frac_comp", 1e1, index=comp)
        # Set scaling for gas phase thermophysical and transport properties
        iron_oc.fs.gas_props.set_default_scaling("enth_mol", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("enth_mol_comp", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("cp_mol", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("cp_mol_comp", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("cp_mass", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("entr_mol", 1e-2)
        iron_oc.fs.gas_props.set_default_scaling("entr_mol_phase", 1e-2)
        iron_oc.fs.gas_props.set_default_scaling("dens_mol", 1)
        iron_oc.fs.gas_props.set_default_scaling("dens_mol_comp", 1)
        iron_oc.fs.gas_props.set_default_scaling("dens_mass", 1e2)
        iron_oc.fs.gas_props.set_default_scaling("visc_d_comp", 1e4)
        iron_oc.fs.gas_props.set_default_scaling("diffusion_comp", 1e5)
        iron_oc.fs.gas_props.set_default_scaling("therm_cond_comp", 1e2)
        iron_oc.fs.gas_props.set_default_scaling("visc_d", 1e5)
        iron_oc.fs.gas_props.set_default_scaling("therm_cond", 1e0)
        iron_oc.fs.gas_props.set_default_scaling("mw", 1e2)

        # Set scaling for solid phase state variables
        iron_oc.fs.solid_props.set_default_scaling("flow_mass", 1e-3)
        iron_oc.fs.solid_props.set_default_scaling("particle_porosity", 1e2)
        iron_oc.fs.solid_props.set_default_scaling("temperature", 1e-2)
        for comp in iron_oc.fs.solid_props.component_list:
            iron_oc.fs.solid_props.set_default_scaling(
                "mass_frac_comp", 1e1, index=comp
            )

        # Set scaling for solid phase thermophysical and transport properties
        iron_oc.fs.solid_props.set_default_scaling("enth_mass", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("enth_mol_comp", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("cp_mol_comp", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("cp_mass", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("dens_mass_particle", 1e-2)
        iron_oc.fs.solid_props.set_default_scaling("dens_mass_skeletal", 1e-2)

        FB0D = iron_oc.fs.unit  # alias to keep test lines short

        # Calculate scaling factors
        iscale.calculate_scaling_factors(FB0D)

        assert pytest.approx(0.10000, rel=1e-5) == iscale.get_scaling_factor(
            FB0D.bed_diameter
        )
        assert pytest.approx(0.10000, rel=1e-5) == iscale.get_scaling_factor(
            FB0D.bed_height
        )
        assert pytest.approx(0.127324, rel=1e-5) == iscale.get_scaling_factor(
            FB0D.volume_bed
        )

        for c in FB0D.volume_bed_constraint.values():
            assert pytest.approx(
                0.127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.volume_solid_constraint.items():
            assert pytest.approx(
                0.127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in FB0D.solids_material_holdup_constraints.items():
            assert pytest.approx(
                0.00127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in FB0D.solids_material_accumulation_constraints.items():
            assert pytest.approx(
                0.127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.mass_solids_constraint.items():
            assert pytest.approx(
                0.00127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.sum_component_constraint.items():
            assert pytest.approx(
                10.00000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.solids_energy_holdup_constraints.items():
            assert pytest.approx(
                1.27324e-7, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.solids_energy_accumulation_constraints.items():
            assert pytest.approx(
                1.27324e-7, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        optarg = {"tol": 1e-6}

        initialization_tester(iron_oc, optarg=optarg)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_by_time(self, iron_oc):

        optarg = {"tol": 1e-6}
        solver = get_solver("ipopt", optarg)  # create solver

        initialize_by_time_element(iron_oc.fs, iron_oc.fs.time, solver=solver)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        assert iron_oc.fs.unit.solids[iron_oc.fs.time.first()].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solids[t].particle_porosity.fixed
            assert iron_oc.fs.unit.solids[t].density_particle_constraint.active
            assert iron_oc.fs.unit.solids[t].density_skeletal_constraint.active

        # Assert that constraints are feasible after initialization
        for con in iron_oc.fs.component_data_objects(Constraint, active=True):
            assert value(con.body) - value(con.upper) < 1e-4
            assert value(con.lower) - value(con.body) < 1e-4

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):
        assert (
            pytest.approx(1798.8532, rel=1e-5)
            == iron_oc.fs.unit.mass_solids[3600].value
        )
        assert (
            pytest.approx(0.195476, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].particle_porosity.value
        )
        assert (
            pytest.approx(0.558299, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].mass_frac_comp["Al2O3"].value
        )
        assert (
            pytest.approx(0.00511114, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].mass_frac_comp["Fe2O3"].value
        )
        assert (
            pytest.approx(0.43659, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].mass_frac_comp["Fe3O4"].value
        )
        assert (
            pytest.approx(1255.59, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].temperature.value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        dt, T = 36, 3600  # size of time interval and time domain
        # first, check if Final Mass - Initial Mass = Total Amount Reacted
        mbal_solid_total = value(
            (iron_oc.fs.unit.mass_solids[T] - iron_oc.fs.unit.mass_solids[0])
            - (
                sum(
                    iron_oc.fs.unit.reactions[t].reaction_rate[r]
                    * iron_oc.fs.unit.config.reaction_package.rate_reaction_stoichiometry[
                        r, "Sol", j
                    ]
                    * iron_oc.fs.unit.solids[t]._params.mw_comp[j]
                    * iron_oc.fs.unit.volume_solid[t]
                    * dt
                    for t in list(iron_oc.fs.time)[1:]
                    for j in ("Al2O3", "Fe2O3", "Fe3O4")
                    for r in iron_oc.fs.unit.config.reaction_package.rate_reaction_idx
                )
            )
        )
        assert abs(mbal_solid_total) <= 1e-2

        # second, check if Change in Mass = Amount Reacted at each time t
        mbal_solid = []
        for t in list(iron_oc.fs.time)[1:]:  # skip t = 0
            mbal_solid.append(
                value(
                    (
                        iron_oc.fs.unit.mass_solids[t]
                        - iron_oc.fs.unit.mass_solids[t - dt]
                    )
                    - (
                        sum(
                            iron_oc.fs.unit.reactions[t].reaction_rate[r]
                            * iron_oc.fs.unit.config.reaction_package.rate_reaction_stoichiometry[
                                r, "Sol", j
                            ]
                            * iron_oc.fs.unit.solids[t]._params.mw_comp[j]
                            * iron_oc.fs.unit.volume_solid[t]
                            * dt
                            for j in ("Al2O3", "Fe2O3", "Fe3O4")
                            for r in iron_oc.fs.unit.config.reaction_package.rate_reaction_idx
                        )
                    )
                )
            )
        for val in mbal_solid:
            assert abs(val) <= 1e-2

        # Reaction stoichiometric ratio (gas is fixed, can only check solids)
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        fe2o3_reacted = value(
            (
                iron_oc.fs.unit.mass_solids[0]
                * iron_oc.fs.unit.solids[0].mass_frac_comp["Fe2O3"]
                - iron_oc.fs.unit.mass_solids[T]
                * iron_oc.fs.unit.solids[T].mass_frac_comp["Fe2O3"]
            )
            / iron_oc.fs.unit.solids[0]._params.mw_comp["Fe2O3"]
        )
        fe3o4_produced = value(
            (
                iron_oc.fs.unit.mass_solids[T]
                * iron_oc.fs.unit.solids[T].mass_frac_comp["Fe3O4"]
                - iron_oc.fs.unit.mass_solids[0]
                * iron_oc.fs.unit.solids[0].mass_frac_comp["Fe3O4"]
            )
            / iron_oc.fs.unit.solids[0]._params.mw_comp["Fe3O4"]
        )
        stoichiometric_ratio = fe2o3_reacted / fe3o4_produced
        assert pytest.approx(1.5, rel=1e-5) == stoichiometric_ratio

        # Conservation of energy check
        # first, check if Final Energy - Initial Energy = Reaction Enthalpy
        ebal_solid_total = value(
            (
                iron_oc.fs.unit.mass_solids[T] * iron_oc.fs.unit.solids[T].enth_mass
                - iron_oc.fs.unit.mass_solids[0] * iron_oc.fs.unit.solids[0].enth_mass
            )
            - (
                -sum(
                    iron_oc.fs.unit.reactions[t].reaction_rate[r]
                    * iron_oc.fs.unit.reactions[t].dh_rxn[r]
                    * iron_oc.fs.unit.volume_solid[t]
                    * dt
                    for t in list(iron_oc.fs.time)[1:]
                    for r in iron_oc.fs.unit.config.reaction_package.rate_reaction_idx
                )
            )
        )
        assert abs(ebal_solid_total) <= 1e-2

        # second, check if Change in Energy = Change in Holdup at each time t
        ebal_solid = []
        for t in list(iron_oc.fs.time)[1:]:  # skip t = 0
            ebal_solid.append(
                value(
                    (
                        iron_oc.fs.unit.mass_solids[t]
                        * iron_oc.fs.unit.solids[t].enth_mass
                        - iron_oc.fs.unit.mass_solids[t - dt]
                        * iron_oc.fs.unit.solids[t - dt].enth_mass
                    )
                    - (
                        -sum(
                            iron_oc.fs.unit.reactions[t].reaction_rate[r]
                            * iron_oc.fs.unit.reactions[t].dh_rxn[r]
                            * iron_oc.fs.unit.volume_solid[t]
                            * dt
                            for r in iron_oc.fs.unit.config.reaction_package.rate_reaction_idx
                        )
                    )
                )
            )
        for val in ebal_solid:
            assert abs(val) <= 1e-2

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        # no inlet/outlet ports (no stream table), test block reports instead
        iron_oc.fs.unit.gas.report()
        iron_oc.fs.unit.solids.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.volume_bed.fix(1)

        with pytest.raises(InitializationError):
            optarg = {"tol": 1e-6}
            iron_oc.fs.unit.initialize(optarg=optarg)


# -----------------------------------------------------------------------------
class TestIronOC_EnergyBalanceType(object):
    @pytest.fixture(scope="class")
    def iron_oc(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 3600], time_units=pyunits.s)

        m.fs.gas_props = GasPhaseParameterBlock()
        m.fs.solid_props = SolidPhaseParameterBlock()
        m.fs.solid_rxns = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
        )

        m.fs.unit = FixedBed0D(
            energy_balance_type=EnergyBalanceType.none,
            gas_property_package=m.fs.gas_props,
            solid_property_package=m.fs.solid_props,
            reaction_package=m.fs.solid_rxns,
        )

        # Discretize time domain
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=100, wrt=m.fs.time, scheme="BACKWARD")

        # Set reactor design conditions
        m.fs.unit.bed_diameter.fix(1)  # diameter of the TGA reactor [m]
        m.fs.unit.bed_height.fix(1)  # height of solids in the TGA reactor [m]

        # Set initial conditions of the solid phase
        m.fs.unit.solids[0].temperature.fix(1273.15)
        m.fs.unit.solids[0].particle_porosity.fix(0.20)
        m.fs.unit.solids[0].mass_frac_comp["Fe2O3"].fix(0.45)
        m.fs.unit.solids[0].mass_frac_comp["Fe3O4"].fix(0)
        m.fs.unit.solids[0].mass_frac_comp["Al2O3"].fix(0.55)

        # Set conditions of the gas phase (this is all fixed as gas side
        # assumption is excess gas flowrate which means all state variables
        # remain unchanged)
        for t in m.fs.time:
            m.fs.unit.gas[t].temperature.fix(1273.15)
            m.fs.unit.gas[t].pressure.fix(1.01325e5)  # 1atm
            m.fs.unit.gas[t].mole_frac_comp["CO2"].fix(0.4)
            m.fs.unit.gas[t].mole_frac_comp["H2O"].fix(0.5)
            m.fs.unit.gas[t].mole_frac_comp["CH4"].fix(0.1)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iron_oc):
        assert isinstance(iron_oc.fs.unit.volume_bed_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.volume_solid_constraint, Constraint)
        assert isinstance(
            iron_oc.fs.unit.solids_material_holdup_constraints, Constraint
        )
        assert isinstance(
            iron_oc.fs.unit.solids_material_accumulation_constraints, Constraint
        )
        assert isinstance(iron_oc.fs.unit.mass_solids_constraint, Constraint)
        assert isinstance(iron_oc.fs.unit.sum_component_constraint, Constraint)

        assert number_variables(iron_oc) == 3044
        assert number_total_constraints(iron_oc) == 2319
        assert number_unused_variables(iron_oc) == 206
        print(unused_variables_set(iron_oc))

    @pytest.mark.unit
    def test_dof(self, iron_oc):
        assert degrees_of_freedom(iron_oc) == 0

    @pytest.fixture(scope="class")
    def iron_oc_unscaled(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 3600], time_units=pyunits.s)

        m.fs.gas_props = GasPhaseParameterBlock()
        m.fs.solid_props = SolidPhaseParameterBlock()
        m.fs.solid_rxns = HeteroReactionParameterBlock(
            solid_property_package=m.fs.solid_props, gas_property_package=m.fs.gas_props
        )

        m.fs.unit = FixedBed0D(
            energy_balance_type=EnergyBalanceType.none,
            gas_property_package=m.fs.gas_props,
            solid_property_package=m.fs.solid_props,
            reaction_package=m.fs.solid_rxns,
        )

        # Discretize time domain
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=100, wrt=m.fs.time, scheme="BACKWARD")

        # Set reactor design conditions
        m.fs.unit.bed_diameter.fix(1)  # diameter of the TGA reactor [m]
        m.fs.unit.bed_height.fix(1)  # height of solids in the TGA reactor [m]

        # Set initial conditions of the solid phase
        m.fs.unit.solids[0].temperature.fix(1273.15)
        m.fs.unit.solids[0].particle_porosity.fix(0.20)
        m.fs.unit.solids[0].mass_frac_comp["Fe2O3"].fix(0.45)
        m.fs.unit.solids[0].mass_frac_comp["Fe3O4"].fix(0)
        m.fs.unit.solids[0].mass_frac_comp["Al2O3"].fix(0.55)

        # Set conditions of the gas phase (this is all fixed as gas side
        # assumption is excess gas flowrate which means all state variables
        # remain unchanged)
        for t in m.fs.time:
            m.fs.unit.gas[t].temperature.fix(1273.15)
            m.fs.unit.gas[t].pressure.fix(1.01325e5)  # 1atm
            m.fs.unit.gas[t].mole_frac_comp["CO2"].fix(0.4)
            m.fs.unit.gas[t].mole_frac_comp["H2O"].fix(0.5)
            m.fs.unit.gas[t].mole_frac_comp["CH4"].fix(0.1)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_unscaled(self, iron_oc_unscaled):
        optarg = {"tol": 1e-6}

        initialization_tester(iron_oc_unscaled, optarg=optarg)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_by_time_unscaled(self, iron_oc_unscaled):

        optarg = {"tol": 1e-6}
        solver = get_solver("ipopt", optarg)  # create solver

        initialize_by_time_element(
            iron_oc_unscaled.fs, iron_oc_unscaled.fs.time, solver=solver
        )

        assert degrees_of_freedom(iron_oc_unscaled) == 0

        # Assert that model is still fixed and deactivated as expected
        assert iron_oc_unscaled.fs.unit.solids[
            iron_oc_unscaled.fs.time.first()
        ].particle_porosity.fixed

        for t in iron_oc_unscaled.fs.time:
            if t != iron_oc_unscaled.fs.time.first():
                assert not iron_oc_unscaled.fs.unit.solids[t].particle_porosity.fixed
            assert iron_oc_unscaled.fs.unit.solids[t].density_particle_constraint.active
            assert iron_oc_unscaled.fs.unit.solids[t].density_skeletal_constraint.active

        # Assert that constraints are feasible after initialization
        for con in iron_oc_unscaled.fs.component_data_objects(Constraint, active=True):
            assert value(con.body) - value(con.upper) < 1e-4
            assert value(con.lower) - value(con.body) < 1e-4

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_unscaled(self, iron_oc_unscaled):
        results = solver.solve(iron_oc_unscaled)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_scaling(self, iron_oc):

        # Set scaling gas phase for state variables
        iron_oc.fs.gas_props.set_default_scaling("flow_mol", 1e-3)
        iron_oc.fs.gas_props.set_default_scaling("pressure", 1e-5)
        iron_oc.fs.gas_props.set_default_scaling("temperature", 1e-2)
        for comp in iron_oc.fs.gas_props.component_list:
            iron_oc.fs.gas_props.set_default_scaling("mole_frac_comp", 1e1, index=comp)
        # Set scaling for gas phase thermophysical and transport properties
        iron_oc.fs.gas_props.set_default_scaling("enth_mol", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("enth_mol_comp", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("cp_mol", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("cp_mol_comp", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("cp_mass", 1e-6)
        iron_oc.fs.gas_props.set_default_scaling("entr_mol", 1e-2)
        iron_oc.fs.gas_props.set_default_scaling("entr_mol_phase", 1e-2)
        iron_oc.fs.gas_props.set_default_scaling("dens_mol", 1)
        iron_oc.fs.gas_props.set_default_scaling("dens_mol_comp", 1)
        iron_oc.fs.gas_props.set_default_scaling("dens_mass", 1e2)
        iron_oc.fs.gas_props.set_default_scaling("visc_d_comp", 1e4)
        iron_oc.fs.gas_props.set_default_scaling("diffusion_comp", 1e5)
        iron_oc.fs.gas_props.set_default_scaling("therm_cond_comp", 1e2)
        iron_oc.fs.gas_props.set_default_scaling("visc_d", 1e5)
        iron_oc.fs.gas_props.set_default_scaling("therm_cond", 1e0)
        iron_oc.fs.gas_props.set_default_scaling("mw", 1e2)

        # Set scaling for solid phase state variables
        iron_oc.fs.solid_props.set_default_scaling("flow_mass", 1e-3)
        iron_oc.fs.solid_props.set_default_scaling("particle_porosity", 1e2)
        iron_oc.fs.solid_props.set_default_scaling("temperature", 1e-2)
        for comp in iron_oc.fs.solid_props.component_list:
            iron_oc.fs.solid_props.set_default_scaling(
                "mass_frac_comp", 1e1, index=comp
            )

        # Set scaling for solid phase thermophysical and transport properties
        iron_oc.fs.solid_props.set_default_scaling("enth_mass", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("enth_mol_comp", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("cp_mol_comp", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("cp_mass", 1e-6)
        iron_oc.fs.solid_props.set_default_scaling("dens_mass_particle", 1e-2)
        iron_oc.fs.solid_props.set_default_scaling("dens_mass_skeletal", 1e-2)

        FB0D = iron_oc.fs.unit  # alias to keep test lines short

        # Calculate scaling factors
        iscale.calculate_scaling_factors(FB0D)

        assert pytest.approx(0.10000, rel=1e-5) == iscale.get_scaling_factor(
            FB0D.bed_diameter
        )
        assert pytest.approx(0.10000, rel=1e-5) == iscale.get_scaling_factor(
            FB0D.bed_height
        )
        assert pytest.approx(0.127324, rel=1e-5) == iscale.get_scaling_factor(
            FB0D.volume_bed
        )

        for c in FB0D.volume_bed_constraint.values():
            assert pytest.approx(
                0.127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.volume_solid_constraint.items():
            assert pytest.approx(
                0.127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in FB0D.solids_material_holdup_constraints.items():
            assert pytest.approx(
                0.00127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for (t, j), c in FB0D.solids_material_accumulation_constraints.items():
            assert pytest.approx(
                0.127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.mass_solids_constraint.items():
            assert pytest.approx(
                0.00127324, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.sum_component_constraint.items():
            assert pytest.approx(
                10.00000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)
        for t, c in FB0D.isothermal_solid_phase.items():
            assert pytest.approx(
                0.01000, rel=1e-5
            ) == iscale.get_constraint_transform_applied_scaling_factor(c)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iron_oc):
        optarg = {"tol": 1e-6}

        initialization_tester(iron_oc, optarg=optarg)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_by_time(self, iron_oc):

        optarg = {"tol": 1e-6}
        solver = get_solver("ipopt", optarg)  # create solver

        initialize_by_time_element(iron_oc.fs, iron_oc.fs.time, solver=solver)

        assert degrees_of_freedom(iron_oc) == 0

        # Assert that model is still fixed and deactivated as expected
        assert iron_oc.fs.unit.solids[iron_oc.fs.time.first()].particle_porosity.fixed

        for t in iron_oc.fs.time:
            if t != iron_oc.fs.time.first():
                assert not iron_oc.fs.unit.solids[t].particle_porosity.fixed
            assert iron_oc.fs.unit.solids[t].density_particle_constraint.active
            assert iron_oc.fs.unit.solids[t].density_skeletal_constraint.active

        # Assert that constraints are feasible after initialization
        for con in iron_oc.fs.component_data_objects(Constraint, active=True):
            assert value(con.body) - value(con.upper) < 1e-4
            assert value(con.lower) - value(con.body) < 1e-4

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iron_oc):
        results = solver.solve(iron_oc)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iron_oc):  # need to update these values
        assert (
            pytest.approx(1798.8281, rel=1e-5)
            == iron_oc.fs.unit.mass_solids[3600].value
        )
        assert (
            pytest.approx(0.195472, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].particle_porosity.value
        )
        assert (
            pytest.approx(0.558307, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].mass_frac_comp["Al2O3"].value
        )
        assert (
            pytest.approx(0.00469410, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].mass_frac_comp["Fe2O3"].value
        )
        assert (
            pytest.approx(0.4370, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].mass_frac_comp["Fe3O4"].value
        )
        assert (
            pytest.approx(1273.15, rel=1e-5)
            == iron_oc.fs.unit.solids[3600].temperature.value
        )

    @pytest.mark.component
    def test_units_consistent(self, iron_oc):
        assert_units_consistent(iron_oc)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iron_oc):
        # Conservation of material check
        dt, T = 36, 3600  # size of time interval and time domain
        # first, check if Final Mass - Initial Mass = Total Amount Reacted
        mbal_solid_total = value(
            (iron_oc.fs.unit.mass_solids[T] - iron_oc.fs.unit.mass_solids[0])
            - (
                sum(
                    iron_oc.fs.unit.reactions[t].reaction_rate[r]
                    * iron_oc.fs.unit.config.reaction_package.rate_reaction_stoichiometry[
                        r, "Sol", j
                    ]
                    * iron_oc.fs.unit.solids[t]._params.mw_comp[j]
                    * iron_oc.fs.unit.volume_solid[t]
                    * dt
                    for t in list(iron_oc.fs.time)[1:]
                    for j in ("Al2O3", "Fe2O3", "Fe3O4")
                    for r in iron_oc.fs.unit.config.reaction_package.rate_reaction_idx
                )
            )
        )
        assert abs(mbal_solid_total) <= 1e-2

        # second, check if Change in Mass = Amount Reacted at each time t
        mbal_solid = []
        for t in list(iron_oc.fs.time)[1:]:  # skip t = 0
            mbal_solid.append(
                value(
                    (
                        iron_oc.fs.unit.mass_solids[t]
                        - iron_oc.fs.unit.mass_solids[t - dt]
                    )
                    - (
                        sum(
                            iron_oc.fs.unit.reactions[t].reaction_rate[r]
                            * iron_oc.fs.unit.config.reaction_package.rate_reaction_stoichiometry[
                                r, "Sol", j
                            ]
                            * iron_oc.fs.unit.solids[t]._params.mw_comp[j]
                            * iron_oc.fs.unit.volume_solid[t]
                            * dt
                            for j in ("Al2O3", "Fe2O3", "Fe3O4")
                            for r in iron_oc.fs.unit.config.reaction_package.rate_reaction_idx
                        )
                    )
                )
            )
        for val in mbal_solid:
            assert abs(val) <= 1e-2

        # Reaction stoichiometric ratio (gas is fixed, can only check solids)
        # Overall reducer reactions for methane combustion:
        # CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O
        fe2o3_reacted = value(
            (
                iron_oc.fs.unit.mass_solids[0]
                * iron_oc.fs.unit.solids[0].mass_frac_comp["Fe2O3"]
                - iron_oc.fs.unit.mass_solids[T]
                * iron_oc.fs.unit.solids[T].mass_frac_comp["Fe2O3"]
            )
            / iron_oc.fs.unit.solids[0]._params.mw_comp["Fe2O3"]
        )
        fe3o4_produced = value(
            (
                iron_oc.fs.unit.mass_solids[T]
                * iron_oc.fs.unit.solids[T].mass_frac_comp["Fe3O4"]
                - iron_oc.fs.unit.mass_solids[0]
                * iron_oc.fs.unit.solids[0].mass_frac_comp["Fe3O4"]
            )
            / iron_oc.fs.unit.solids[0]._params.mw_comp["Fe3O4"]
        )
        stoichiometric_ratio = fe2o3_reacted / fe3o4_produced
        assert pytest.approx(1.5, rel=1e-5) == stoichiometric_ratio

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, iron_oc):
        # no inlet/outlet ports (no stream table), test block reports instead
        iron_oc.fs.unit.gas.report()
        iron_oc.fs.unit.solids.report()

    @pytest.mark.component
    def test_initialization_error(self, iron_oc):
        iron_oc.fs.unit.volume_bed.fix(1)

        with pytest.raises(InitializationError):
            optarg = {"tol": 1e-6}
            iron_oc.fs.unit.initialize(optarg=optarg)
